from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tox.pytest import ToxProjectCreator


def test_extra_setup_commands_runs_with_notest(tox_project: ToxProjectCreator) -> None:
    ini = """
        [testenv]
        package = skip
        deps = pip
        extra_setup_commands = python -c 'print("extra setup")'
        commands = python -c 'print("main command")'
    """
    proj = tox_project({"tox.ini": ini})
    result = proj.run("r", "--notest")
    result.assert_success()
    assert "extra setup" in result.out
    assert "main command" not in result.out


def test_extra_setup_commands_runs_without_notest(tox_project: ToxProjectCreator) -> None:
    ini = """
        [testenv]
        package = skip
        deps = pip
        extra_setup_commands = python -c 'print("extra setup")'
        commands = python -c 'print("main command")'
    """
    proj = tox_project({"tox.ini": ini})
    result = proj.run("r")
    result.assert_success()
    assert "extra setup" in result.out
    assert "main command" in result.out


def test_extra_setup_commands_with_package_skip(tox_project: ToxProjectCreator) -> None:
    ini = """
        [testenv]
        package = skip
        extra_setup_commands = python -c 'import sys; sys.exit(0)'
    """
    proj = tox_project({"tox.ini": ini})
    result = proj.run("r", "--notest")
    result.assert_success()


def test_extra_setup_commands_with_package_install(tox_project: ToxProjectCreator) -> None:
    ini = "[testenv]\nextra_setup_commands = python -c 'print(\"after package install\")'"
    proj = tox_project({"tox.ini": ini})
    result = proj.run("r", "--notest")
    result.assert_success()
    assert "after package install" in result.out


def test_extra_setup_commands_failure_stops_execution(tox_project: ToxProjectCreator) -> None:
    ini = """
        [testenv]
        package = skip
        extra_setup_commands = python -c 'import sys; sys.exit(1)'
        commands = python -c 'print("should not run")'
    """
    proj = tox_project({"tox.ini": ini})
    result = proj.run("r", "--notest")
    result.assert_failed()
    assert "extra_setup_commands failed" in result.out


def test_extra_setup_commands_failure_with_ignore_errors(tox_project: ToxProjectCreator) -> None:
    ini = """
        [testenv]
        package = skip
        ignore_errors = true
        extra_setup_commands = python -c 'import sys; sys.exit(1)'
        commands = python -c 'print("main command")'
    """
    proj = tox_project({"tox.ini": ini})
    result = proj.run("r")
    result.assert_success()
    assert "main command" in result.out


def test_extra_setup_commands_multiple_commands(tox_project: ToxProjectCreator) -> None:
    ini = """
        [testenv]
        package = skip
        extra_setup_commands =
            python -c 'print("cmd1")'
            python -c 'print("cmd2")'
            python -c 'print("cmd3")'
    """
    proj = tox_project({"tox.ini": ini})
    result = proj.run("r", "--notest")
    result.assert_success()
    assert "cmd1" in result.out
    assert "cmd2" in result.out
    assert "cmd3" in result.out


def test_extra_setup_commands_empty_default(tox_project: ToxProjectCreator) -> None:
    ini = "[testenv]\npackage = skip"
    proj = tox_project({"tox.ini": ini})
    result = proj.run("r", "--notest")
    result.assert_success()


def test_extra_setup_commands_execution_order(tox_project: ToxProjectCreator) -> None:
    ini = """
        [testenv]
        package = skip
        deps = pip
        extra_setup_commands = python -c 'print("EXTRA_SETUP")'
        commands_pre = python -c 'print("COMMANDS_PRE")'
        commands = python -c 'print("COMMANDS")'
        commands_post = python -c 'print("COMMANDS_POST")'
    """
    proj = tox_project({"tox.ini": ini})
    result = proj.run("r")
    result.assert_success()

    assert "EXTRA_SETUP" in result.out
    assert "COMMANDS_PRE" in result.out
    assert "COMMANDS" in result.out
    assert "COMMANDS_POST" in result.out

    lines = result.out.split("\n")
    extra_setup_idx = next(i for i, line in enumerate(lines) if "EXTRA_SETUP" in line)
    commands_pre_idx = next(i for i, line in enumerate(lines) if "COMMANDS_PRE" in line)
    commands_idx = next(
        i
        for i, line in enumerate(lines)
        if "COMMANDS" in line and "COMMANDS_PRE" not in line and "COMMANDS_POST" not in line
    )
    commands_post_idx = next(i for i, line in enumerate(lines) if "COMMANDS_POST" in line)

    assert extra_setup_idx < commands_pre_idx < commands_idx < commands_post_idx



from __future__ import annotations

import math
from pathlib import Path, PurePosixPath

import pytest

from tox.config.loader.native import to_native
from tox.config.set_env import SetEnv
from tox.config.types import Command, EnvList
from tox.tox_env.python.pip.req_file import PythonDeps


def test_str() -> None:
    assert to_native("hello") == "hello"


def test_bool_true() -> None:
    assert to_native(True) is True


def test_bool_false() -> None:
    assert to_native(False) is False


def test_bool_before_int() -> None:
    result = to_native(True)
    assert result is True
    assert not isinstance(result, int) or isinstance(result, bool)


def test_int() -> None:
    assert to_native(42) == 42


def test_float() -> None:
    assert to_native(math.pi) == math.pi


def test_path(tmp_path: Path) -> None:
    assert to_native(tmp_path) == str(tmp_path)


def test_posix_path() -> None:
    assert to_native(PurePosixPath("/usr/bin")) == "/usr/bin"


def test_dict() -> None:
    assert to_native({"a": 1, "b": "c"}) == {"a": 1, "b": "c"}


def test_nested_dict() -> None:
    assert to_native({"a": {"b": 1}}) == {"a": {"b": 1}}


def test_list() -> None:
    assert to_native([1, "two", 3.0]) == [1, "two", 3.0]


def test_list_of_paths(tmp_path: Path) -> None:
    p = tmp_path / "a"
    assert to_native([p]) == [str(p)]


def test_set() -> None:
    result = to_native({"c", "a", "b"})
    assert result == ["a", "b", "c"]


def test_env_list() -> None:
    assert to_native(EnvList(["py39", "py310", "lint"])) == ["py39", "py310", "lint"]


def test_command_simple() -> None:
    assert to_native(Command(["pytest"])) == "pytest"


@pytest.mark.parametrize(
    ("args", "expected_prefix"),
    [
        (["pytest", "-v"], "pytest"),
        (["-", "cmd"], "cmd"),
        (["!", "cmd"], "cmd"),
    ],
)
def test_command_variants(args: list[str], expected_prefix: str) -> None:
    result = to_native(Command(args))
    assert isinstance(result, str)
    assert expected_prefix in result


def test_set_env(tmp_path: Path) -> None:
    raw = "A=1\nB=hello"
    se = SetEnv(raw, "set_env", "py", tmp_path)
    result = to_native(se)
    assert result == {"A": "1", "B": "hello"}


def test_python_deps(tmp_path: Path) -> None:
    (tmp_path / "tox.ini").touch()
    deps = PythonDeps("pytest\nflask>=2.0", tmp_path)
    result = to_native(deps)
    assert result == ["pytest", "flask>=2.0"]


def test_fallback_unknown_type() -> None:
    class Custom:
        def __str__(self) -> str:
            return "custom-value"

    assert to_native(Custom()) == "custom-value"


def test_empty_str() -> None:
    assert to_native("") is not None
    assert isinstance(to_native(""), str)
    assert len(to_native("")) == 0


def test_zero() -> None:
    assert to_native(0) == 0
    assert isinstance(to_native(0), int)


def test_float_zero() -> None:
    result = to_native(0.0)
    assert isinstance(result, float)
    assert result == pytest.approx(0.0)



from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path
from typing import TYPE_CHECKING

from tox.plugin import impl

if TYPE_CHECKING:
    from tox.config.cli.parser import ToxParser
    from tox.session.state import State


@impl
def tox_add_option(parser: ToxParser) -> None:
    from tox.config.cli.parser import CORE  # noqa: PLC0415

    parser.add_command("man", [], "Set up tox man page for current shell", setup_man, inherit=frozenset({CORE}))


def setup_man(state: State) -> int:  # noqa: ARG001
    print("tox man page setup")  # noqa: T201
    print("=" * 50)  # noqa: T201
    print()  # noqa: T201

    man_in_wheel = Path(sys.prefix) / "share" / "man" / "man1" / "tox.1"
    print(f"Looking for man page at: {man_in_wheel}")  # noqa: T201
    print()  # noqa: T201

    if not man_in_wheel.exists():
        print("✗ Man page not found")  # noqa: T201
        print("  The man page should be included in the wheel when tox is installed.")  # noqa: T201
        print("  If you installed from source or in development mode, the man page")  # noqa: T201
        print("  may not be present. Try installing from PyPI or use 'tox --help'.")  # noqa: T201
        return 1

    print("✓ Man page found")  # noqa: T201
    print()  # noqa: T201

    if _check_man_accessible():
        print("✓ 'man tox' already works!")  # noqa: T201
        print("  No setup needed.")  # noqa: T201
        return 0

    if (exit_code := _create_symlink(man_in_wheel)) != 0:
        return exit_code

    print()  # noqa: T201

    manpath = os.environ.get("MANPATH", "")
    user_man_base = str(Path.home() / ".local" / "share" / "man")

    if user_man_base in manpath:
        print("✓ MANPATH already includes ~/.local/share/man")  # noqa: T201
        print("  Try running: man tox")  # noqa: T201
        return 0

    print("⚠ MANPATH does not include ~/.local/share/man")  # noqa: T201
    print()  # noqa: T201
    _print_manpath_instructions()

    return 0


def _check_man_accessible() -> bool:
    try:
        result = subprocess.run(["man", "tox"], capture_output=True, text=True, timeout=2, check=False)  # noqa: S607
    except (subprocess.TimeoutExpired, FileNotFoundError):
        return False
    else:
        return result.returncode == 0


def _create_symlink(man_in_wheel: Path) -> int:
    user_man_dir = Path.home() / ".local" / "share" / "man" / "man1"
    user_man_file = user_man_dir / "tox.1"

    user_man_dir.mkdir(parents=True, exist_ok=True)

    if user_man_file.exists() or user_man_file.is_symlink():
        if user_man_file.is_symlink() and user_man_file.resolve() == man_in_wheel:
            print(f"✓ Symlink already exists: {user_man_file} → {man_in_wheel}")  # noqa: T201
            return 0
        print(f"✗ File already exists: {user_man_file}")  # noqa: T201
        print("  Remove it manually if you want to replace it.")  # noqa: T201
        return 1

    user_man_file.symlink_to(man_in_wheel)
    print(f"✓ Created symlink: {user_man_file} → {man_in_wheel}")  # noqa: T201
    return 0


def _print_manpath_instructions() -> None:
    shell = os.environ.get("SHELL", "")
    is_fish = "fish" in shell

    rc_file = {
        True: "~/.config/fish/config.fish",
        "bash" in shell: "~/.bashrc",
        "zsh" in shell: "~/.zshrc",
    }.get(is_fish or any(s in shell for s in ("bash", "zsh")), "~/.profile")

    print(f"To complete setup, add this to {rc_file}:")  # noqa: T201
    print()  # noqa: T201

    export_line = (
        'set -x MANPATH "$HOME/.local/share/man" $MANPATH'
        if is_fish
        else 'export MANPATH="$HOME/.local/share/man:$MANPATH"'
    )
    print(f"  {export_line}")  # noqa: T201
    print()  # noqa: T201
    print("Then restart your shell or run:")  # noqa: T201
    print(f"  source {rc_file}")  # noqa: T201
    print()  # noqa: T201
    print("After that, you can use: man tox")  # noqa: T201



"""On Windows we use overlapped I/O for efficient real-time stream reading."""

from __future__ import annotations  # pragma: win32 cover

import _overlapped  # type: ignore[import-untyped]  # pragma: win32 cover # noqa: PLC2701
import time  # pragma: win32 cover
from typing import TYPE_CHECKING

from .read_via_thread import ReadViaThread  # pragma: win32 cover

if TYPE_CHECKING:
    from collections.abc import Callable

READ_CHUNK_SIZE = 32768  # pragma: win32 cover
POLL_INTERVAL = 0.05  # pragma: win32 cover
ERROR_IO_INCOMPLETE = 996  # pragma: win32 cover


class ReadViaThreadWindows(ReadViaThread):  # pragma: win32 cover
    def __init__(self, file_no: int, handler: Callable[[bytes], int], name: str, drain: bool) -> None:  # noqa: FBT001
        super().__init__(file_no, handler, name, drain)

    def _read_stream(self) -> None:
        try:
            while not self.stop.is_set():
                ov = _overlapped.Overlapped(0)
                try:
                    ov.ReadFile(self.file_no, READ_CHUNK_SIZE)
                except OSError:
                    break

                while True:
                    try:
                        data = ov.getresult(False)  # noqa: FBT003
                        break
                    except OSError as exception:
                        if getattr(exception, "winerror", None) != ERROR_IO_INCOMPLETE:
                            return
                        time.sleep(POLL_INTERVAL)

                if not data:
                    break
                self.handler(data)
        except OSError:  # pragma: no cover
            pass

    def _drain_stream(self) -> None:
        try:
            while True:
                ov = _overlapped.Overlapped(0)
                try:
                    ov.ReadFile(self.file_no, READ_CHUNK_SIZE)
                except OSError:
                    break

                try:
                    data = ov.getresult(True)  # noqa: FBT003
                except OSError:
                    break

                if not data:
                    break
                self.handler(data)
        except OSError:  # pragma: no cover
            pass



from __future__ import annotations

import sys
from argparse import Action
from typing import TYPE_CHECKING

import pytest

from tox.config.cli.parser import Parsed, ToxParser

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from tox.pytest import CaptureFixture, MonkeyPatch


def test_parser_const_with_default_none(monkeypatch: MonkeyPatch) -> None:
    monkeypatch.setenv("TOX_ALPHA", "2")
    parser = ToxParser.base()
    parser.add_argument(
        "-a",
        dest="alpha",
        action="store_const",
        const=1,
        default=None,
        help="sum the integers (default: find the max)",
    )
    parser.fix_defaults()

    result = parser.parse_args([])
    assert result.alpha == 2


@pytest.mark.parametrize("is_atty", [True, False])
@pytest.mark.parametrize("no_color", [None, "0", "1", "", "\t", " ", "false", "true"])
@pytest.mark.parametrize("force_color", [None, "", "0", "1", "always", "not an empty string"])
@pytest.mark.parametrize("tox_color", [None, "bad", "no", "yes"])
@pytest.mark.parametrize("tty_compatible", [None, "", "0", "1", "other"])
@pytest.mark.parametrize("term", [None, "xterm", "dumb"])
def test_parser_color(  # noqa: PLR0913
    monkeypatch: MonkeyPatch,
    mocker: MockerFixture,
    no_color: str | None,
    force_color: str | None,
    tox_color: str | None,
    tty_compatible: str | None,
    is_atty: bool,
    term: str | None,
) -> None:
    for key, value in {
        "NO_COLOR": no_color,
        "TOX_COLORED": tox_color,
        "FORCE_COLOR": force_color,
        "TTY_COMPATIBLE": tty_compatible,
        "TERM": term,
    }.items():
        if value is None:
            monkeypatch.delenv(key, raising=False)
        else:
            monkeypatch.setenv(key, value)
    stdout_mock = mocker.patch("tox.config.cli.parser.sys.stdout")
    stdout_mock.isatty.return_value = is_atty

    if tox_color in {"yes", "no"}:
        expected = tox_color == "yes"
    elif bool(no_color):
        expected = False
    elif bool(force_color):
        expected = True
    elif tty_compatible in {"0", "1"}:
        expected = tty_compatible == "1"
    elif term == "dumb":
        expected = False
    else:
        expected = is_atty

    is_colored = ToxParser.base().parse_args([], Parsed()).is_colored
    assert is_colored is expected


def test_parser_unsupported_type() -> None:
    parser = ToxParser.base()
    parser.add_argument("--magic", action="store", default=None)
    with pytest.raises(TypeError) as context:
        parser.fix_defaults()
    action = context.value.args[0]
    assert isinstance(action, Action)
    assert action.dest == "magic"


def test_sub_sub_command() -> None:
    parser = ToxParser.base()
    with pytest.raises(RuntimeError, match="no sub-command group allowed"):
        parser.add_command(
            "c",
            [],
            "help",
            lambda s: 0,  # noqa: ARG005
        )  # pragma: no cover - the lambda will never be run


def test_parse_known_args_not_set(mocker: MockerFixture) -> None:
    mocker.patch.object(sys, "argv", ["a", "--help"])
    parser = ToxParser.base()
    _, unknown = parser.parse_known_args(None)
    assert unknown == ["--help"]


def test_parser_hint(capsys: CaptureFixture) -> None:
    parser = ToxParser.base()
    with pytest.raises(SystemExit):
        parser.parse_args("foo")
    _out, err = capsys.readouterr()
    assert err.endswith("hint: if you tried to pass arguments to a command use -- to separate them from tox ones\n")



from __future__ import annotations

import textwrap
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tox.pytest import ToxProjectCreator


def test_env_base_simple_factors(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.task]
            factors = ["x", "y"]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    assert "task-x" in result.out
    assert "task-y" in result.out


def test_env_base_cartesian_product(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.task]
            factors = [["a", "b"], ["x", "y"]]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    for env in ("task-a-x", "task-a-y", "task-b-x", "task-b-y"):
        assert env in result.out


def test_env_base_range_factors(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.task]
            factors = [{"prefix" = "py3", "start" = 12, "stop" = 13}]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    assert "task-py312" in result.out
    assert "task-py313" in result.out


def test_env_base_config_inheritance(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.lib]
            factors = ["a", "b"]
            package = "skip"
            description = "from env_base"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "lib-a", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:lib-a]\ndescription = from env_base\n", "")


def test_env_base_inherits_from_env_run_base(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_run_base]
            description = "from run_base"

            [env_base.lib]
            factors = ["a"]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "lib-a", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:lib-a]\ndescription = from run_base\n", "")


def test_env_base_factor_conditionals(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.django]
            factors = [["a", "b"], ["py314", "py315"]]
            package = "skip"
            description = {replace = "if", condition = "factor.py315", then = "uses315", else = "uses314"}
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "django-a-py314", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:django-a-py314]\ndescription = uses314\n", "")
    outcome = project.run("c", "-e", "django-b-py315", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:django-b-py315]\ndescription = uses315\n", "")


def test_env_base_template_not_listed(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.lib]
            factors = ["a", "b"]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    lines = result.out.splitlines()
    env_names = [line.split()[0] for line in lines if line.strip() and " -> " in line]
    assert "lib" not in env_names


def test_env_base_explicit_env_override(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.lib]
            factors = ["a", "b"]
            package = "skip"
            description = "from base"
            commands = [["python", "-c", "print('ok')"]]

            [env.lib-a]
            description = "overridden"
        """),
    })
    outcome = project.run("c", "-e", "lib-a", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:lib-a]\ndescription = overridden\n", "")
    outcome = project.run("c", "-e", "lib-b", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:lib-b]\ndescription = from base\n", "")


def test_env_base_multiple_entries(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.lib]
            factors = ["a"]
            package = "skip"
            commands = [["python", "-c", "print('lib')"]]

            [env_base.app]
            factors = ["x"]
            package = "skip"
            commands = [["python", "-c", "print('app')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    assert "lib-a" in result.out
    assert "app-x" in result.out


def test_env_base_no_unused_warnings(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.lib]
            factors = ["a"]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "lib-a")
    outcome.assert_success()
    assert "# !!! unused: " not in outcome.out


def test_env_base_generated_envs_can_run(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.task]
            factors = ["a"]
            package = "skip"
            commands = [["python", "-c", "print('hello')"]]
        """),
    })
    result = project.run("r", "-e", "task-a")
    result.assert_success()
    assert "hello" in result.out


def test_env_base_pyproject_toml(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "pyproject.toml": textwrap.dedent("""\
            [tool.tox.env_base.lib]
            factors = ["a", "b"]
            package = "skip"
            description = "pyproject"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "lib-a", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:lib-a]\ndescription = pyproject\n", "")


def test_env_base_missing_factors(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.lib]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_failed()
    assert "env_base.lib requires a 'factors' key; use [env.lib] for single environments" in result.out


def test_env_base_factors_not_a_list(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.lib]
            factors = "bad"
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_failed()
    assert "env_base.lib.factors must be a list" in result.out


def test_env_base_env_run_base_override_order(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_run_base]
            description = "from run_base"

            [env_base.lib]
            factors = ["a"]
            package = "skip"
            description = "from env_base"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "lib-a", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:lib-a]\ndescription = from env_base\n", "")


def test_env_base_core_no_unused_warning(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.lib]
            factors = ["a"]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "--core")
    outcome.assert_success()
    assert "# !!! unused: " not in outcome.out


def test_env_base_deps_from_template(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.lib]
            factors = ["a"]
            package = "skip"
            deps = ["pytest>=8", "coverage"]
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "lib-a", "-k", "deps")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:lib-a]\ndeps =\n  pytest>=8\n  coverage\n", "")


def test_env_base_doc_getting_started_scaling(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.test]
            factors = [["3.13", "3.14"]]
            deps = ["pytest>=8"]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    assert "test-3.13" in result.out
    assert "test-3.14" in result.out


def test_env_base_doc_reference_generative(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.build]
            factors = [["py312", "py313"], ["x86", "x64"]]
            package = "skip"
            env_dir = {replace = "if", condition = "factor.x86", then = ".venv-x86", else = ".venv-x64"}
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    for env in ("build-py312-x86", "build-py312-x64", "build-py313-x86", "build-py313-x64"):
        assert env in result.out


def test_env_base_doc_reference_django_matrix(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.django]
            factors = [["py312", "py313"], ["django42", "django50"]]
            package = "skip"
            deps = [
                "pytest",
                {replace = "if", condition = "factor.django42", then = ["Django>=4.2,<4.3"]},
                {replace = "if", condition = "factor.django50", then = ["Django>=5.0,<5.1"]},
            ]
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    for env in ("django-py312-django42", "django-py312-django50", "django-py313-django42", "django-py313-django50"):
        assert env in result.out


def test_env_base_doc_howto_matrix(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.django]
            factors = [
                {"prefix" = "py3", "start" = 13, "stop" = 14},
                ["django42", "django50"],
            ]
            package = "skip"
            deps = [
                "pytest",
                {replace = "if", condition = "factor.django42", then = ["Django>=4.2,<4.3"]},
                {replace = "if", condition = "factor.django50", then = ["Django>=5.0,<5.1"]},
            ]
            commands = [["pytest"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    for env in ("django-py313-django42", "django-py313-django50", "django-py314-django42", "django-py314-django50"):
        assert env in result.out


def test_env_base_doc_howto_override(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.django]
            factors = [
                {"prefix" = "py3", "start" = 13, "stop" = 14},
                ["django42", "django50"],
            ]
            package = "skip"
            deps = ["pytest"]
            commands = [["python", "-c", "print('ok')"]]

            [env.django-py314-django50]
            description = "bleeding edge"
        """),
    })
    outcome = project.run("c", "-e", "django-py314-django50", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:django-py314-django50]\ndescription = bleeding edge\n", "")


def test_env_base_doc_reference_generative_section_names(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.py311-venv]
            factors = [["x86", "x64"]]
            package = "skip"
            base_python = {replace = "if", condition = "factor.x86", then = "python3.11-32", else = "python3.11-64"}
            env_dir = {replace = "if", condition = "factor.x86", then = ".venv-x86", else = ".venv-x64"}
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    assert "py311-venv-x86" in result.out
    assert "py311-venv-x64" in result.out


def test_env_base_cartesian_with_range(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.django]
            factors = [["a", "b"], {"prefix" = "py3", "start" = 12, "stop" = 13}]
            package = "skip"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    result = project.run("l")
    result.assert_success()
    for env in ("django-a-py312", "django-a-py313", "django-b-py312", "django-b-py313"):
        assert env in result.out


def test_env_base_keyed_factor_description(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.sync]
            factors = [{ecosystem = ["oci", "python"]}]
            package = "skip"
            description = "Sync {factor:ecosystem} artifacts"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "sync-oci", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:sync-oci]\ndescription = Sync oci artifacts\n", "")
    outcome = project.run("c", "-e", "sync-python", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:sync-python]\ndescription = Sync python artifacts\n", "")


def test_env_base_keyed_factor_two_groups(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.sync]
            factors = [{ecosystem = ["oci", "python"]}, {target = ["pw", "tt"]}]
            package = "skip"
            description = "Sync {factor:ecosystem} to {factor:target}"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "sync-oci-pw", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:sync-oci-pw]\ndescription = Sync oci to pw\n", "")
    outcome = project.run("c", "-e", "sync-python-tt", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:sync-python-tt]\ndescription = Sync python to tt\n", "")


def test_env_base_positional_factor_description(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.sync]
            factors = [["oci", "python"], ["pw", "tt"]]
            package = "skip"
            description = "Sync {factor:0} to {factor:1}"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "sync-oci-pw", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:sync-oci-pw]\ndescription = Sync oci to pw\n", "")


def test_env_base_keyed_factor_mixed_with_plain(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.task]
            factors = [["a", "b"], {target = ["x", "y"]}]
            package = "skip"
            description = "{factor:0} on {factor:target}"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "task-a-x", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:task-a-x]\ndescription = a on x\n", "")


def test_env_base_keyed_factor_in_commands(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.sync]
            factors = [{ecosystem = ["oci", "python"]}]
            package = "skip"
            commands = [["python", "-c", "print('{factor:ecosystem}')"]]
        """),
    })
    result = project.run("r", "-e", "sync-oci")
    result.assert_success()
    assert "oci" in result.out


def test_env_base_factor_label_no_match_returns_empty(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.task]
            factors = [{ecosystem = ["oci", "python"]}]
            package = "skip"
            description = "Got [{factor:ecosystem}]"
            commands = [["python", "-c", "print('ok')"]]

            [env.other]
            description = "No match [{factor:ecosystem}]"
        """),
    })
    outcome = project.run("c", "-e", "other", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:other]\ndescription = No match []\n", "")


def test_env_base_factor_label_with_default(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": textwrap.dedent("""\
            [env_base.task]
            factors = [{ecosystem = ["oci", "python"]}]
            package = "skip"
            description = "Type is {factor:unknown:fallback}"
            commands = [["python", "-c", "print('ok')"]]
        """),
    })
    outcome = project.run("c", "-e", "task-oci", "-k", "description")
    outcome.assert_success()
    outcome.assert_out_err("[testenv:task-oci]\ndescription = Type is fallback\n", "")



from __future__ import annotations

import logging
import os
from collections import OrderedDict, defaultdict
from itertools import chain, tee
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, cast

from .sets import ConfigSet, CoreConfigSet, EnvConfigSet

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence

    from tox.config.loader.api import Loader, OverrideMap

    from .cli.parser import Parsed
    from .loader.memory import MemoryLoader
    from .loader.section import Section
    from .source import Source


T = TypeVar("T", bound=ConfigSet)


class Config:
    """Main configuration object for tox."""

    def __init__(  # noqa: PLR0913  # <- no way around many args
        self,
        config_source: Source,
        options: Parsed,
        root: Path,
        pos_args: Sequence[str] | None,
        work_dir: Path,
        extra_envs: Iterable[str],
    ) -> None:
        self._pos_args = None if pos_args is None else tuple(pos_args)
        self._work_dir = work_dir
        self._root = root
        self._options = options
        self._extra_envs = extra_envs

        self._overrides: OverrideMap = defaultdict(list)
        for override in options.override:
            self._overrides[override.namespace].append(override)

        self._src = config_source
        self._key_to_conf_set: dict[tuple[str, str, str], ConfigSet] = OrderedDict()
        self._env_to_conf_set: dict[str, EnvConfigSet] = {}
        self._core_set: CoreConfigSet | None = None
        self.memory_seed_loaders: defaultdict[str, list[MemoryLoader]] = defaultdict(list)

    def pos_args(self, to_path: Path | None) -> tuple[str, ...] | None:
        """:param to_path: if not None rewrite relative posargs paths from cwd to to_path

        :returns: positional argument

        """
        if self._pos_args is not None and to_path is not None and Path.cwd() != to_path:
            args = []
            # we use os.path to unroll .. in path without resolve
            to_path_str = os.path.abspath(str(to_path))  # noqa: PTH100
            for arg in self._pos_args:
                path_arg = Path(arg)
                try:
                    is_existing_path = path_arg.exists()
                except OSError:
                    logging.debug("could not check if %r is an existing path", arg)
                    is_existing_path = False
                if is_existing_path and not path_arg.is_absolute():
                    # we use os.path to unroll .. in path without resolve
                    path_arg_str = os.path.abspath(str(path_arg))  # noqa: PTH100
                    try:
                        relative = os.path.relpath(path_arg_str, to_path_str)
                    except ValueError:  # on Windows, relpath fails across drives (e.g. subst mounts)
                        args.append(path_arg_str)
                    else:
                        args.append(relative)
                else:
                    args.append(arg)
            return tuple(args)
        return self._pos_args

    @property
    def work_dir(self) -> Path:
        """:returns: working directory for this project"""
        return self._work_dir

    @property
    def src_path(self) -> Path:
        """:returns: the location of the tox configuration source"""
        return self._src.path

    def __iter__(self) -> Iterator[str]:
        """:returns: an iterator that goes through existing environments"""
        # NOTE: `tee(self._extra_envs)[1]` is necessary for compatibility with
        # NOTE: Python 3.11 and older versions. Once Python 3.12 is the lowest
        # NOTE: supported version, it can be changed to
        # NOTE: `chain.from_iterable(tee(self._extra_envs, 1))`.
        return chain(self._src.envs(self.core), tee(self._extra_envs)[1])

    def sections(self) -> Iterator[Section]:
        yield from self._src.sections()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(config_source={self._src!r})"

    def __contains__(self, item: str) -> bool:
        """:returns: check if an environment already exists"""
        return any(name for name in self if name == item)

    @classmethod
    def make(cls, parsed: Parsed, pos_args: Sequence[str] | None, source: Source, extra_envs: Iterable[str]) -> Config:
        """Make a tox configuration object."""
        # root is the project root, where the configuration file is at
        # work dir is where we put our own files
        root: Path = source.path.parent if parsed.root_dir is None else parsed.root_dir
        work_dir: Path = source.path.parent if parsed.work_dir is None else parsed.work_dir
        # if these are relative we need to expand them them to ensure paths built on this can resolve independent on cwd
        root = root.resolve()
        work_dir = work_dir.resolve()
        return cls(
            config_source=source,
            options=parsed,
            pos_args=pos_args,
            root=root,
            work_dir=work_dir,
            extra_envs=extra_envs,
        )

    @property
    def options(self) -> Parsed:
        return self._options

    @property
    def factor_labels(self) -> dict[str, list[str]]:
        return getattr(self._src, "_factor_labels", {})

    @property
    def core(self) -> CoreConfigSet:
        """:returns: the core configuration"""
        if self._core_set is not None:
            return self._core_set
        core_section = self._src.get_core_section()
        core = CoreConfigSet(self, core_section, self._root, self.src_path)
        core.loaders.extend(self._src.get_loaders(core_section, base=[], override_map=self._overrides, conf=core))
        self._core_set = core
        return core

    def get_section_config(
        self,
        section: Section,
        base: list[str] | None,
        of_type: type[T],
        for_env: str | None,
        loaders: Sequence[Loader[Any]] | None = None,
    ) -> T:
        key = section.key, for_env or "", "-".join(base or [])
        try:
            return cast("T", self._key_to_conf_set[key])
        except KeyError:
            conf_set = of_type(self, section, for_env)
            self._key_to_conf_set[key] = conf_set
            if for_env is not None:
                conf_set.loaders.extend(self.memory_seed_loaders.get(for_env, []))
            for loader in self._src.get_loaders(section, base, self._overrides, conf_set):
                conf_set.loaders.append(loader)
            if loaders is not None:
                conf_set.loaders.extend(loaders)
            return conf_set

    def get_env(
        self,
        item: str,
        package: bool = False,  # noqa: FBT001, FBT002
        loaders: Sequence[Loader[Any]] | None = None,
    ) -> EnvConfigSet:
        """Return the configuration for a given tox environment (will create if not exist yet).

        :param item: the name of the environment is
        :param package: a flag indicating if the environment is of type packaging or not (only used for creation)
        :param loaders: loaders to use for this configuration (only used for creation)

        :returns: the tox environments config

        """
        if item in self._env_to_conf_set:
            return self._env_to_conf_set[item]
        section, base_test, base_pkg = self._src.get_tox_env_section(item)
        result = self.get_section_config(
            section,
            base=base_pkg if package else base_test,
            of_type=EnvConfigSet,
            for_env=item,
            loaders=loaders,
        )
        self._env_to_conf_set[item] = result
        return result

    def clear_env(self, name: str) -> None:
        section, _, __ = self._src.get_tox_env_section(name)
        self._key_to_conf_set = {
            k: v for k, v in self._key_to_conf_set.items() if not (k[0] == section.key and k[1] == name)
        }
        self._env_to_conf_set.pop(name, None)


___all__ = [
    "Config",
]



"""Subprocess-based virtualenv session/creator for use with pinned virtualenv versions."""

from __future__ import annotations

import hashlib
import json
import logging
import subprocess
import sys
import venv
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from filelock import FileLock

if TYPE_CHECKING:
    from pathlib import Path


_PROBE_SCRIPT = """\
import json, struct, sys, sysconfig
print(json.dumps({
    "implementation": sys.implementation.name,
    "version_info": list(sys.version_info[:5]),
    "version": sys.version.split()[0],
    "architecture": struct.calcsize("P") * 8,
    "platform": sys.platform,
    "system_executable": sys.executable,
    "free_threaded": sysconfig.get_config_var("Py_GIL_DISABLED") == 1,
    "sysconfig_platform": sysconfig.get_platform(),
}))
"""


@dataclass
class _VersionInfo:
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int


@dataclass
class SubprocessPythonInfo:
    """Python interpreter information gathered via subprocess probe."""

    implementation: str
    version_info: _VersionInfo
    version: str
    architecture: int
    platform: str
    system_executable: str
    free_threaded: bool
    sysconfig_platform: str | None = None

    @property
    def machine(self) -> str:
        """Derive instruction set architecture from sysconfig_platform."""
        if (plat := self.sysconfig_platform) is None:
            return ""
        parts = plat.rsplit("-", 1)
        return parts[-1] if len(parts) > 1 else ""


@dataclass
class SubprocessCreator:
    """Mimics virtualenv Creator + Describe interface using known venv directory layout."""

    _env_dir: Path
    interpreter: SubprocessPythonInfo
    _is_win: bool = field(default_factory=lambda: sys.platform == "win32")

    @property
    def bin_dir(self) -> Path:
        return self._env_dir / ("Scripts" if self._is_win else "bin")

    @property
    def script_dir(self) -> Path:
        return self.bin_dir

    @property
    def purelib(self) -> Path:
        if self._is_win:
            return self._env_dir / "Lib" / "site-packages"
        vi = self.interpreter.version_info
        return self._env_dir / "lib" / f"python{vi.major}.{vi.minor}" / "site-packages"

    @property
    def platlib(self) -> Path:
        return self.purelib

    @property
    def exe(self) -> Path:
        return self.bin_dir / ("python.exe" if self._is_win else "python")


class SubprocessSession:
    """Mimics virtualenv Session interface, runs a bootstrapped virtualenv via subprocess."""

    def __init__(
        self,
        env_dir: Path,
        bootstrap_python: Path,
        env_vars: dict[str, str],
        interpreter: SubprocessPythonInfo | None,
    ) -> None:
        self._env_dir = env_dir
        self._bootstrap_python = bootstrap_python
        self._env_vars = env_vars
        self._creator = SubprocessCreator(env_dir, interpreter) if interpreter is not None else None

    def run(self) -> None:
        cmd = [str(self._bootstrap_python), "-m", "virtualenv", str(self._env_dir)]
        try:
            result = subprocess.run(cmd, env=self._env_vars, capture_output=True, text=True, check=False)
        except FileNotFoundError as exc:
            msg = f"virtualenv subprocess failed: {exc}"
            raise RuntimeError(msg) from exc
        if result.returncode != 0:
            msg = f"virtualenv subprocess failed (exit {result.returncode}): {result.stderr}"
            raise RuntimeError(msg)

    @property
    def creator(self) -> SubprocessCreator:
        if self._creator is None:
            msg = "no interpreter discovered"
            raise RuntimeError(msg)
        return self._creator


def probe_python(python_path: str) -> SubprocessPythonInfo | None:
    """Probe a Python executable to extract interpreter metadata."""
    try:
        result = subprocess.run(
            [python_path, "-c", _PROBE_SCRIPT], capture_output=True, text=True, timeout=30, check=False
        )
        if result.returncode != 0:
            return None
        raw = json.loads(result.stdout)
    except (subprocess.SubprocessError, json.JSONDecodeError, FileNotFoundError, OSError):
        return None
    vi = raw["version_info"]
    return SubprocessPythonInfo(
        implementation=raw["implementation"],
        version_info=_VersionInfo(major=vi[0], minor=vi[1], micro=vi[2], releaselevel=vi[3], serial=vi[4]),
        version=raw["version"],
        architecture=raw["architecture"],
        platform=raw["platform"],
        system_executable=raw["system_executable"],
        free_threaded=raw["free_threaded"],
        sysconfig_platform=raw.get("sysconfig_platform"),
    )


def _bootstrap_path(work_dir: Path, virtualenv_spec: str) -> Path:
    digest = hashlib.sha256(virtualenv_spec.encode()).hexdigest()[:16]
    return work_dir / ".virtualenv-bootstrap" / digest


def _bin_dir(base: Path) -> Path:
    return base / ("Scripts" if sys.platform == "win32" else "bin")


def _bootstrap_python(base: Path) -> Path:
    return _bin_dir(base) / ("python.exe" if sys.platform == "win32" else "python")


def _bootstrap_pip(base: Path) -> Path:
    return _bin_dir(base) / ("pip.exe" if sys.platform == "win32" else "pip")


def _has_correct_virtualenv(python: Path, virtualenv_spec: str) -> bool:
    if not python.exists():
        return False
    try:
        result = subprocess.run(
            [str(python), "-c", "from importlib.metadata import version; print(version('virtualenv'))"],
            capture_output=True,
            text=True,
            timeout=30,
            check=False,
        )
        if result.returncode != 0:
            return False
        from packaging.specifiers import SpecifierSet  # noqa: PLC0415

        installed = result.stdout.strip()
        spec = virtualenv_spec.removeprefix("virtualenv")
        return installed in SpecifierSet(spec) if spec else True
    except (subprocess.SubprocessError, FileNotFoundError, OSError):
        return False


def ensure_bootstrap(work_dir: Path, virtualenv_spec: str) -> Path:
    """Create or reuse a cached bootstrap venv with the specified virtualenv version."""
    base = _bootstrap_path(work_dir, virtualenv_spec)
    python = _bootstrap_python(base)
    if _has_correct_virtualenv(python, virtualenv_spec):
        return python

    lock_path = base.parent / f"{base.name}.lock"
    lock_path.parent.mkdir(parents=True, exist_ok=True)
    with FileLock(lock_path):
        if _has_correct_virtualenv(python, virtualenv_spec):
            return python
        logging.info("bootstrapping %s into %s", virtualenv_spec, base)
        if base.exists():
            import shutil  # noqa: PLC0415

            shutil.rmtree(base)
        venv.create(str(base), with_pip=True, clear=True)
        pip = _bootstrap_pip(base)
        result = subprocess.run([str(pip), "install", virtualenv_spec], capture_output=True, text=True, check=False)
        if result.returncode != 0:
            msg = f"failed to install {virtualenv_spec} into bootstrap env: {result.stderr}"
            raise RuntimeError(msg)
        return python



from __future__ import annotations

import importlib
import sys
from pathlib import Path
from textwrap import dedent
from types import ModuleType
from typing import TYPE_CHECKING, Any, Optional, TypeVar, Union

import pytest

from tox.config.loader.str_convert import StrConvert
from tox.config.types import Command, EnvList

if TYPE_CHECKING:
    from tox.pytest import MonkeyPatch, SubRequest, ToxProjectCreator

from typing import Literal


@pytest.mark.parametrize(
    ("raw", "value", "of_type"),
    [
        pytest.param("true", True, bool, id="str_true_to_bool"),
        pytest.param("false", False, bool, id="str_false_to_bool"),
        pytest.param("True", True, bool, id="str_True_to_bool"),
        pytest.param("False", False, bool, id="str_False_to_bool"),
        pytest.param("TruE", True, bool, id="str_TruE_to_bool"),
        pytest.param("FalsE", False, bool, id="str_FalsE_to_bool"),
        pytest.param("1", True, bool, id="str_1_to_bool"),
        pytest.param("0", False, bool, id="str_0_to_bool"),
        pytest.param("1", 1, int, id="str_1_to_int"),
        pytest.param("0", 0, int, id="str_0_to_int"),
        pytest.param("+1", 1, int, id="str_plus1_to_int"),
        pytest.param("-1", -1, int, id="str_minus1_to_int"),
        pytest.param("1.1", 1.1, float, id="str_1.1_to_float"),
        pytest.param("0.1", 0.1, float, id="str_0.1_to_float"),
        pytest.param("+1.1", 1.1, float, id="str_plus1.1_to_float"),
        pytest.param("-1.1", -1.1, float, id="str_minus1.1_to_float"),
        pytest.param("magic", "magic", str, id="str_magic_to_str"),
        pytest.param("1", {"1"}, set[str], id="str_1_to_set_str"),
        pytest.param("1", [1], list[int], id="str_1_to_list_int"),
        pytest.param("1=2", {1: 2}, dict[int, int], id="str_1eq2_to_dict_int_int"),
        pytest.param("a=1\n\nc=2", {"a": 1, "c": 2}, dict[str, int], id="str_multiline_dict_str_int"),
        pytest.param("a", Path("a"), Path, id="str_a_to_path"),
        pytest.param("a", Command(["a"]), Command, id="str_a_to_command"),
        pytest.param("a,b", EnvList(["a", "b"]), EnvList, id="str_a_b_to_envlist"),
        pytest.param("", None, Optional[int], id="empty_to_optional_int"),  # noqa: UP045
        pytest.param("1", 1, Optional[int], id="str_1_to_optional_int"),  # noqa: UP045
        pytest.param("", None, Optional[str], id="empty_to_optional_str"),  # noqa: UP045
        pytest.param("1", "1", Optional[str], id="str_1_to_optional_str"),  # noqa: UP045
        pytest.param("", None, Optional[list[str]], id="empty_to_optional_list_str"),  # noqa: UP045
        pytest.param("1,2", ["1", "2"], Optional[list[str]], id="str_1_2_to_optional_list_str"),  # noqa: UP045
        pytest.param("", None, int | None, id="empty_to_int_or_none"),
        pytest.param("1", 1, int | None, id="str_1_to_int_or_none"),
        pytest.param("", None, str | None, id="empty_to_str_or_none"),
        pytest.param("1", "1", str | None, id="str_1_to_str_or_none"),
        pytest.param("", None, list[str] | None, id="empty_to_list_str_or_none"),
        pytest.param("1,2", ["1", "2"], list[str] | None, id="str_1_2_to_list_str_or_none"),
        pytest.param("1", "1", Literal["1", "2"], id="str_1_to_literal_1_2"),
    ],
)
def test_str_convert_ok(raw: str, value: Any, of_type: type[Any]) -> None:
    result = StrConvert().to(raw, of_type, None)
    assert result == value


@pytest.mark.parametrize(
    ("raw", "of_type", "exc_type", "msg"),
    [
        pytest.param("a", TypeVar, TypeError, r"a cannot cast to .*typing.TypeVar.*", id="fail_typevar"),
        pytest.param("3", Literal["1", "2"], ValueError, r"3 must be one of \('1', '2'\)", id="fail_literal_1_2"),
        pytest.param(
            "3",
            Union[str, int],  # noqa: UP007
            TypeError,
            r"3 cannot cast to (typing.Union\[str, int\]|str \| int)",
            id="fail_union_str_int",
        ),
        pytest.param(
            "3",
            str | int,
            TypeError,
            r"3 cannot cast to (typing.Union\[str, int\]|str \| int)",
            id="fail_union_bar_str_int",
        ),
        pytest.param("", Command, ValueError, r"attempting to parse '' into a command failed", id="fail_empty_command"),
    ],
)
def test_str_convert_nok(raw: str, of_type: type[Any], msg: str, exc_type: type[Exception]) -> None:
    with pytest.raises(exc_type, match=msg):
        StrConvert().to(raw, of_type, None)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        ("python ' ok", ["python", "' ok"]),
        ('python " ok', ["python", '" ok']),
    ],
)
def test_invalid_shell_expression(value: str, expected: list[str]) -> None:
    result = StrConvert().to_command(value)
    assert result is not None
    assert result.args == expected


SIMPLE_ARGS = [
    ('foo "bar baz"', ["foo", "bar baz"]),
    ('foo "bar baz"ext', ["foo", "bar bazext"]),
    ('foo="bar baz"', ["foo=bar baz"]),
    ("foo 'bar baz'", ["foo", "bar baz"]),
    ("foo 'bar baz'ext", ["foo", "bar bazext"]),
    ("foo='bar baz'", ["foo=bar baz"]),
    (r"foo=\"bar baz\"", ['foo="bar', 'baz"']),
    (r'foo="bar baz\"', ['foo="bar baz\\"']),
    ("foo='bar baz' quuc", ["foo=bar baz", "quuc"]),
    (r"foo='bar baz\' quuc", ["foo=bar baz\\", "quuc"]),
    (r"foo=\"bar baz\' quuc", ['foo="bar', "baz'", "quuc"]),
    (r"foo=\\\"bar baz\"", ['foo=\\"bar', 'baz"']),
    (r'foo=\\"bar baz\"', [r'foo=\\"bar baz\"']),
]
NEWLINE_ARGS = [
    ('foo\n"bar\nbaz"', ["foo", "bar\nbaz"]),
]
INI_CONFIG_NEWLINE_ARGS = [
    ('foo\\\n    "bar\\\n    baz"', ["foobarbaz"]),  # behavior change from tox 3
    ('foo\\\n    "bar \\\n    baz"', ["foobar baz"]),  # behavior change from tox 3
    ('foo \\\n    "bar\\\n    baz"', ["foo", "barbaz"]),
    ('foo \\\n    "bar \\\n    baz"', ["foo", "bar baz"]),
    ('foo \\\n    \\"bar \\\n    baz"', ["foo", '"bar', 'baz"']),
    ("foo \\\n    bar \\\n    baz", ["foo", "bar", "baz"]),
]
WINDOWS_PATH_ARGS = [
    (r"SPECIAL:\foo\bar --quuz='baz atan'", [r"SPECIAL:\foo\bar", "--quuz=baz atan"]),
    (r"X:\\foo\\bar --quuz='baz atan'", [r"X:\foo\bar", "--quuz=baz atan"]),
    ("/foo/bar --quuz='baz atan'", ["/foo/bar", "--quuz=baz atan"]),
    ('cc --arg "C:\\\\Users\\""', ["cc", "--arg", 'C:\\Users"']),
    ('cc --arg "C:\\\\Users\\"', ["cc", "--arg", '"C:\\\\Users\\"']),
    ('cc --arg "C:\\\\Users"', ["cc", "--arg", "C:\\Users"]),
    ('cc --arg \\"C:\\\\Users"', ["cc", "--arg", '\\"C:\\\\Users"']),
    ('cc --arg "C:\\\\Users\\ "', ["cc", "--arg", "C:\\Users\\ "]),
    ('cc --arg "C:\\\\Users\\\\"', ["cc", "--arg", "C:\\Users\\"]),
    ('cc --arg "C:\\\\Users\\\\ "', ["cc", "--arg", "C:\\Users\\ "]),
    (
        r'cc --arg C:\\Users\\ --arg2 "SPECIAL:\Temp\f o o" --arg3="\\FOO\share\Path name" --arg4 SPECIAL:\Temp\ '[:-1],
        [
            "cc",
            "--arg",
            "C:\\Users\\",
            "--arg2",
            "SPECIAL:\\Temp\\f o o",
            "--arg3=\\FOO\\share\\Path name",
            "--arg4",
            "SPECIAL:\\Temp\\",
        ],
    ),
]
WINDOWS_TRAILING_SEP_ARGS = [
    (r"command path\to\trailing\sep\ -b foo", ["command", "path\\to\\trailing\\sep\\", "-b", "foo"]),
]
WACKY_SLASH_ARGS = [
    ("\\\\\\", ["\\\\\\"]),
    (" \\'\\'\\ '", [" \\'\\'\\ '"]),
    ("\\'\\'\\ ", ["'' "]),
    ("\\'\\ \\\\", ["' \\"]),
    ("\\'\\ ", ["' "]),
    ('''"\\'\\"''', ['"\\\'\\"']),
    ("'\\' \\\\", ["\\", "\\"]),
    ('"\\\\" \\\\', ["\\", "\\"]),
]
WACKY_SLASH_ARGS_WIN32 = {
    " \\'\\'\\ '": ["''\\", "'"],
    "\\'\\'\\ ": ["''\\"],
    "\\'\\ \\\\": ["'\\", "\\"],
    "\\'\\ ": ["'\\"],
}


@pytest.fixture(params=["win32", "linux2"])
def sys_platform(request: SubRequest, monkeypatch: MonkeyPatch) -> str:
    class _SelectiveSys(ModuleType):
        """A sys-like proxy that only overrides `platform`."""

        def __init__(self, patched_platform: str) -> None:
            super().__init__("sys")
            self.__dict__["_real"] = sys
            self.__dict__["_patched_platform"] = patched_platform

        def __getattr__(self, name: str) -> Any:
            if name == "platform":
                return self.__dict__["_patched_platform"]
            return getattr(self.__dict__["_real"], name)

    # Patches sys.platform only for the tox.config.loader.str_convert module.
    # Everywhere else, sys.platform remains the real value.
    mod = importlib.import_module("tox.config.loader.str_convert")
    proxy = _SelectiveSys(str(request.param))
    monkeypatch.setattr(mod, "sys", proxy, raising=True)
    return str(request.param)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        *SIMPLE_ARGS,
        *NEWLINE_ARGS,
        *WINDOWS_PATH_ARGS,
        *WACKY_SLASH_ARGS,
    ],
)
def test_shlex_platform_specific(sys_platform: str, value: str, expected: list[str]) -> None:
    if sys_platform != "win32" and value.startswith("SPECIAL:"):
        # on non-windows platform, backslash is always an escape, not path separator
        expected = [exp.replace("\\", "") for exp in expected]
    if sys_platform == "win32" and value in WACKY_SLASH_ARGS_WIN32:
        expected = WACKY_SLASH_ARGS_WIN32[value]
    result = StrConvert().to_command(value)
    assert result is not None
    assert result.args == expected


@pytest.mark.parametrize(("value", "expected"), WINDOWS_TRAILING_SEP_ARGS)
def test_shlex_win32_trailing_sep(sys_platform: str, value: str, expected: list[str]) -> None:
    if sys_platform != "win32":
        pytest.skip("trailing path separator only relevant on Windows")
    result = StrConvert().to_command(value)
    assert result is not None
    assert result.args == expected


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        *SIMPLE_ARGS,
        *INI_CONFIG_NEWLINE_ARGS,
        *WINDOWS_PATH_ARGS,
        #        *WACKY_SLASH_ARGS,
    ],
)
def test_shlex_platform_specific_ini(
    tox_project: ToxProjectCreator,
    sys_platform: str,
    value: str,
    expected: list[str],
) -> None:
    if sys_platform != "win32" and value.startswith("SPECIAL:"):
        # on non-windows platform, backslash is always an escape, not path separator
        expected = [exp.replace("\\", "") for exp in expected]
    project = tox_project(
        {
            "tox.ini": dedent(
                """
                [testenv]
                commands =
                    %s""",
            )
            % value,
        },
    )
    outcome = project.run("c")
    outcome.assert_success()
    env_config = outcome.env_conf("py")
    result = env_config["commands"]
    assert result == [Command(args=expected)]


def test_trailing_path_sep_does_not_merge_args(
    tox_project: ToxProjectCreator,
    sys_platform: str,
    monkeypatch: MonkeyPatch,
) -> None:
    if sys_platform != "win32":
        pytest.skip("trailing path separator only relevant on Windows")
    monkeypatch.setattr("os.sep", "\\")
    project = tox_project({
        "tox.ini": "[testenv]\ncommands = command path{/}to{/}sep{/} -b foo",
    })
    outcome = project.run("c")
    outcome.assert_success()
    env_config = outcome.env_conf("py")
    result = env_config["commands"]
    assert result == [Command(args=["command", "path\\to\\sep\\", "-b", "foo"])]



"""Convert string configuration values to tox python configuration objects."""

from __future__ import annotations

import shlex
import sys
from inspect import isclass
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

from tox.config.loader.convert import Convert
from tox.config.types import Command, EnvList

if TYPE_CHECKING:
    from collections.abc import Iterator
    from io import StringIO
    from typing import Final


class StrConvert(Convert[str]):
    """A class converting string values to tox types."""

    @staticmethod
    def to_str(value: str) -> str:
        return str(value).strip()

    @staticmethod
    def to_path(value: str) -> Path:
        return Path(value)

    @staticmethod
    def to_list(value: str, of_type: type[Any]) -> Iterator[str]:
        splitter = "\n" if (isclass(of_type) and issubclass(of_type, Command)) or "\n" in value else ","
        splitter = splitter.replace("\r", "")
        for token in value.split(splitter):
            value = token.strip()
            if value:
                yield value

    @staticmethod
    def to_set(value: str, of_type: type[Any]) -> Iterator[str]:
        yield from StrConvert.to_list(value, of_type)

    @staticmethod
    def to_dict(value: str, of_type: tuple[type[Any], type[Any]]) -> Iterator[tuple[str, str]]:  # noqa: ARG004
        for row in value.split("\n"):
            if row.strip():
                key, sep, value = row.partition("=")
                if sep:
                    yield key.strip(), value.strip()
                else:
                    msg = f"dictionary lines must be of form key=value, found {row!r}"
                    raise TypeError(msg)

    @staticmethod
    def _win32_process_path_backslash(value: str, escape: str, special_chars: str) -> str:
        """Escape backslash in value that is not followed by a special character.

        This allows windows paths to be written without double backslash, while retaining the POSIX backslash escape
        semantics for quotes and escapes.

        """
        result = []
        for ix, char in enumerate(value):
            result.append(char)
            if char == escape:
                last_char = value[ix - 1 : ix]
                if last_char == escape:
                    continue
                next_char = value[ix + 1 : ix + 2]
                if next_char not in {escape, *special_chars}:
                    result.append(escape)  # escape escapes that are not themselves escaping a special character
        return "".join(result)

    @staticmethod
    def to_command(value: str) -> Command | None:
        """At this point, ``value`` has already been substituted out, and all punctuation / escapes are final.

        Value will typically be stripped of whitespace when coming from an ini file.

        """
        value = value.replace(r"\#", "#")
        is_win = sys.platform == "win32"
        if is_win:  # pragma: win32 cover
            s = shlex.shlex(posix=True)
            value = StrConvert._win32_process_path_backslash(
                value,
                escape=s.escape,
                special_chars=s.quotes,
            )
        splitter = shlex.shlex(value, posix=True)
        splitter.whitespace_split = True
        splitter.commenters = ""  # comments handled earlier, and the shlex does not know escaped comment characters
        args: list[str] = []
        pos = 0
        try:
            for arg in splitter:
                if is_win and len(arg) > 1 and arg[0] == arg[-1] and arg.startswith(("'", '"')):  # pragma: win32 cover
                    # on Windows quoted arguments will remain quoted, strip it
                    arg = arg[1:-1]  # noqa: PLW2901
                args.append(arg)
                pos = cast("StringIO", splitter.instream).tell()
        except ValueError:
            args.append(value[pos:])
        if len(args) == 0:
            msg = f"attempting to parse {value!r} into a command failed"
            raise ValueError(msg)
        if args[0] != "-" and args[0].startswith("-"):
            args[0] = args[0][1:]
            args = ["-", *args]
        return Command(args)

    @staticmethod
    def to_env_list(value: str) -> EnvList:
        from tox.config.loader.ini.factor import extend_factors  # noqa: PLC0415

        elements = list(chain.from_iterable(extend_factors(expr) for expr in value.split("\n")))
        return EnvList(elements)

    TRUTHFUL_VALUES: Final[set[str]] = {"true", "1", "yes", "on"}
    FALSE_VALUES: Final[set[str]] = {"false", "0", "no", "off", ""}
    VALID_BOOL = sorted(TRUTHFUL_VALUES | FALSE_VALUES)

    @staticmethod
    def to_bool(value: str) -> bool:
        norm = str(value).strip().lower()
        if norm in StrConvert.TRUTHFUL_VALUES:
            return True
        if norm in StrConvert.FALSE_VALUES:
            return False

        msg = f"value {value!r} cannot be transformed to bool, valid: {', '.join(StrConvert.VALID_BOOL)}"
        raise TypeError(msg)


__all__ = ("StrConvert",)



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from tox.pytest import ToxProjectCreator


def test_retry_succeeds_after_failure(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """\
[env_run_base]
package = "skip"
commands_retry = 3
commands = [["python", "-c", \
"from pathlib import Path; p = Path('counter.txt'); \
c = int(p.read_text()) + 1 if p.exists() else 1; \
p.write_text(str(c)); raise SystemExit(0 if c >= 3 else 1)"]]
"""
    })
    result = proj.run("r", "-e", "py")
    result.assert_success()
    assert "command failed (attempt 1 of 4), retrying" in result.out
    assert "command failed (attempt 2 of 4), retrying" in result.out


def test_retry_all_exhausted(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """\
[env_run_base]
package = "skip"
commands_retry = 2
commands = [["python", "-c", "raise SystemExit(1)"]]
"""
    })
    result = proj.run("r", "-e", "py")
    result.assert_failed(code=1)
    assert "command failed (attempt 1 of 3), retrying" in result.out
    assert "command failed (attempt 2 of 3), retrying" in result.out


def test_retry_default_no_retry(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """\
[env_run_base]
package = "skip"
commands = [["python", "-c", "raise SystemExit(1)"]]
"""
    })
    result = proj.run("r", "-e", "py")
    result.assert_failed(code=1)
    assert "retrying" not in result.out


def test_retry_dash_prefix_no_retry(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """\
[env_run_base]
package = "skip"
commands_retry = 3
commands = [["-", "python", "-c", "raise SystemExit(1)"]]
"""
    })
    result = proj.run("r", "-e", "py")
    result.assert_success()
    assert "retrying" not in result.out


def test_retry_with_ignore_errors(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """\
[env_run_base]
package = "skip"
ignore_errors = true
commands_retry = 1
commands = [["python", "-c", "raise SystemExit(1)"]]
"""
    })
    result = proj.run("r", "-e", "py")
    result.assert_failed(code=1)
    assert "command failed (attempt 1 of 2), retrying" in result.out


def test_retry_bang_prefix_succeeds_on_nonzero(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """\
[env_run_base]
package = "skip"
commands_retry = 2
commands = [["!", "python", "-c", "raise SystemExit(1)"]]
"""
    })
    result = proj.run("r", "-e", "py")
    result.assert_success()
    assert "retrying" not in result.out


def test_retry_bang_prefix_retries_on_zero(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """\
[env_run_base]
package = "skip"
commands_retry = 2
commands = [["!", "python", "-c", \
"from pathlib import Path; p = Path('counter_bang.txt'); \
c = int(p.read_text()) + 1 if p.exists() else 1; \
p.write_text(str(c)); raise SystemExit(0 if c < 3 else 1)"]]
"""
    })
    result = proj.run("r", "-e", "py")
    result.assert_success()
    assert "command failed (attempt 1 of 3), retrying" in result.out
    assert "command failed (attempt 2 of 3), retrying" in result.out


@pytest.mark.parametrize("command_key", ["commands_pre", "commands_post"])
def test_retry_applies_to_pre_and_post(tox_project: ToxProjectCreator, command_key: str) -> None:
    proj = tox_project({
        "tox.toml": f"""\
[env_run_base]
package = "skip"
commands_retry = 3
{command_key} = [["python", "-c", \
"from pathlib import Path; p = Path('counter_{command_key}.txt'); \
c = int(p.read_text()) + 1 if p.exists() else 1; \
p.write_text(str(c)); raise SystemExit(0 if c >= 2 else 1)"]]
"""
    })
    result = proj.run("r", "-e", "py")
    result.assert_success()
    assert "command failed (attempt 1 of 4), retrying" in result.out



from __future__ import annotations

from shutil import rmtree
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

_CACHEDIR_TAG = """\
Signature: 8a477f597d28d172789f06886806bc55
# This file is a cache directory tag created by tox.
# For information about cache directory tags, see:
#	https://bford.info/cachedir/spec.html
"""


def ensure_cachedir_tag(work_dir: Path) -> None:
    """Ensure a ``CACHEDIR.TAG`` file exists in *work_dir* per https://bford.info/cachedir/spec.html."""
    tag_path = work_dir / "CACHEDIR.TAG"
    if not tag_path.exists():
        work_dir.mkdir(parents=True, exist_ok=True)
        tag_path.write_text(_CACHEDIR_TAG, encoding="utf-8")


def ensure_empty_dir(path: Path, except_filename: str | None = None) -> None:
    if path.exists():
        if path.is_dir():
            for sub_path in path.iterdir():
                if sub_path.name == except_filename:
                    continue
                if sub_path.is_dir():
                    rmtree(sub_path, ignore_errors=True)
                else:
                    sub_path.unlink()
        else:
            path.unlink()
            path.mkdir()
    else:
        path.mkdir(parents=True)


def ensure_gitignore(path: Path) -> None:
    """Create a ``.gitignore`` file with ``*`` in the given directory if one does not already exist.

    This prevents tox-managed directories (like ``.tox/``) from being tracked by git, so users don't need to add them to
    their project's ``.gitignore``.

    :param path: the directory in which to create the ``.gitignore`` file

    """
    if not (gitignore := path / ".gitignore").exists():
        path.mkdir(parents=True, exist_ok=True)
        gitignore.write_text("*\n", encoding="utf-8")


__all__ = [
    "ensure_cachedir_tag",
    "ensure_empty_dir",
    "ensure_gitignore",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from tox.execute.util import shebang

if TYPE_CHECKING:
    from pathlib import Path


def test_shebang_found(tmp_path: Path) -> None:
    script_path = tmp_path / "a"
    script_path.write_text("#!  /bin/python \t-c\t")
    assert shebang(str(script_path)) == ["/bin/python", "-c"]


def test_shebang_file_missing(tmp_path: Path) -> None:
    script_path = tmp_path / "a"
    assert shebang(str(script_path)) is None


def test_shebang_no_shebang(tmp_path: Path) -> None:
    script_path = tmp_path / "a"
    script_path.write_bytes(b"magic")
    assert shebang(str(script_path)) is None


def test_shebang_non_utf8_file(tmp_path: Path) -> None:
    script_path, content = tmp_path / "a", b"#!" + bytearray.fromhex("c0")
    script_path.write_bytes(content)
    assert shebang(str(script_path)) is None



"""Expand TOML product dicts into environment name lists."""

from __future__ import annotations

from itertools import product
from typing import Any

from tox.config.loader.ini.factor import LATEST_PYTHON_MINOR_MAX, LATEST_PYTHON_MINOR_MIN


def expand_product(value: dict[str, Any]) -> list[str]:
    """Expand a product dict into a flat list of environment names.

    :param value: dict with ``product`` (list of factor groups) and optional ``exclude`` (list of env names to skip)

    :returns: list of environment names from the cartesian product of all factor groups, joined with ``-``

    """
    raw_groups = value["product"]
    if not isinstance(raw_groups, list):
        msg = f"product value must be a list of factor groups, got {type(raw_groups).__name__}"
        raise TypeError(msg)
    if not raw_groups:
        return []
    expanded = [expand_factor_group(g) for g in raw_groups]
    exclude = set(value.get("exclude") or [])
    return [name for combo in product(*expanded) if (name := "-".join(combo)) not in exclude]


_RESERVED_LABELS: frozenset[str] = frozenset({"env", "posargs", "tty", "glob", "factor"})


def expand_factor_group(group: Any) -> list[str]:
    if isinstance(group, list):
        return [str(item) for item in group]
    if isinstance(group, dict):
        if "prefix" in group:
            return _expand_range(group)
        if len(group) == 1:
            label, values = next(iter(group.items()))
            if label in _RESERVED_LABELS:
                msg = f"'{label}' is reserved and cannot be used as a factor label"
                raise TypeError(msg)
            if not isinstance(values, list):
                msg = f"labeled factor group '{label}' must map to a list, got {type(values).__name__}"
                raise TypeError(msg)
            return [str(v) for v in values]
    msg = f"factor group must be a list, a range dict, or a labeled dict, got {type(group).__name__}"
    raise TypeError(msg)


def extract_label(group: Any) -> str | None:
    if isinstance(group, dict) and "prefix" not in group and len(group) == 1:
        return next(iter(group))
    return None


def _expand_range(range_dict: dict[str, Any]) -> list[str]:
    prefix: str = str(range_dict["prefix"])
    has_start = "start" in range_dict
    has_stop = "stop" in range_dict
    if not has_start and not has_stop:
        msg = "range must have at least 'start' or 'stop'"
        raise TypeError(msg)
    start = range_dict.get("start", LATEST_PYTHON_MINOR_MIN)
    stop = range_dict.get("stop", LATEST_PYTHON_MINOR_MAX)
    if not isinstance(start, int):
        msg = f"range 'start' must be an integer, got {type(start).__name__}"
        raise TypeError(msg)
    if not isinstance(stop, int):
        msg = f"range 'stop' must be an integer, got {type(stop).__name__}"
        raise TypeError(msg)
    return [f"{prefix}{i}" for i in range(start, stop + 1)]


__all__ = [
    "_RESERVED_LABELS",
    "expand_factor_group",
    "expand_product",
    "extract_label",
]



from __future__ import annotations

import re
from collections import OrderedDict
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

import pytest

from tox.config.cli.parser import Parsed
from tox.config.loader.memory import MemoryLoader
from tox.config.main import Config
from tox.config.sets import ConfigSet, EnvConfigSet
from tox.config.source.api import Section

if TYPE_CHECKING:
    from pytest_mock import MockerFixture

    from tests.conftest import ToxIniCreator
    from tox.pytest import ToxProjectCreator

ConfBuilder = Callable[[str], ConfigSet]


@pytest.fixture(name="conf_builder")
def _conf_builder(tox_ini_conf: ToxIniCreator) -> ConfBuilder:
    def _make(conf_str: str) -> ConfigSet:
        return tox_ini_conf(f"[tox]\nenvlist=py39\n[testenv]\n{conf_str}").get_env("py39")

    return _make


def test_config_str(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("deps-x = 1\n    other: 2")
    config_set.add_config(keys="deps-x", of_type=str, default="", desc="ok")
    result = config_set["deps-x"]
    assert result == "1"


def test_config_path(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("path = path")
    config_set.add_config(keys="path", of_type=Path, default=Path(), desc="path")
    path_materialize = config_set["path"]
    assert path_materialize == Path("path")


def test_config_set(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("set = 1\n 2\n 3")
    config_set.add_config(keys="set", of_type=set[int], default=set(), desc="set")
    set_materialize = config_set["set"]
    assert set_materialize == {1, 2, 3}


def test_config_optional_none(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    config_set.add_config(
        keys="optional_none",
        of_type=int | None,
        default=None,
        desc="optional_none",
    )
    optional_none = config_set["optional_none"]
    assert optional_none is None


def test_config_dict(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("dict = a=1\n  b=2\n  c=3")
    config_set.add_config(keys="dict", of_type=dict[str, int], default={}, desc="dict")
    dict_val = config_set["dict"]
    assert dict_val == OrderedDict([("a", 1), ("b", 2), ("c", 3)])


def test_config_bad_type(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("crazy = something-bad")

    config_set.add_config(keys="crazy", of_type=TypeVar, default=TypeVar("V"), desc="crazy")  # ty: ignore[invalid-legacy-type-variable] # intentionally passing TypeVar as type to test error path
    with pytest.raises(TypeError) as context:
        assert config_set["crazy"]
    assert str(context.value) == f"something-bad cannot cast to {TypeVar!r}"


def test_config_bad_dict(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("bad_dict = something")

    config_set.add_config(keys="bad_dict", of_type=dict[str, str], default={}, desc="bad_dict")
    with pytest.raises(TypeError) as context:
        assert config_set["bad_dict"]
    assert str(context.value) == "dictionary lines must be of form key=value, found 'something'"


def test_config_bad_bool(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("bad_bool = whatever")
    config_set.add_config(keys="bad_bool", of_type=bool, default=False, desc="bad_bool")
    with pytest.raises(TypeError) as context:
        assert config_set["bad_bool"]
    error = "value 'whatever' cannot be transformed to bool, valid: , 0, 1, false, no, off, on, true, yes"
    assert str(context.value) == error


def test_config_constant(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    config_set.add_constant(keys="a", value=1, desc="ok")
    const = config_set["a"]
    assert const == 1


def test_config_lazy_constant(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    config_set.add_constant(keys="b", value=lambda: 2, desc="ok")
    lazy_const = config_set["b"]
    assert lazy_const == 2


def test_config_constant_repr(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    defined = config_set.add_constant(keys="a", value=1, desc="ok")
    assert repr(defined)


def test_config_dynamic_repr(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("path = path")
    defined = config_set.add_config(keys="path", of_type=Path, default=Path(), desc="path")
    assert repr(defined)


def test_config_redefine_constant_fail(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("path = path")
    config_set.add_constant(keys="path", desc="desc", value="value")
    msg = (
        "duplicate configuration definition for py39:\n"
        "has: ConfigConstantDefinition(keys=('path',), desc=desc, value=value)\n"
        "new: ConfigConstantDefinition(keys=('path',), desc=desc2, value=value)"
    )
    with pytest.raises(ValueError, match=re.escape(msg)):
        config_set.add_constant(keys="path", desc="desc2", value="value")


def test_config_redefine_dynamic_fail(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("path = path")
    config_set.add_config(keys="path", of_type=str, default="default_1", desc="path")
    msg = (
        "duplicate configuration definition for py39:\n"
        "has: ConfigDynamicDefinition(keys=('path',), desc=path, of_type=<class 'str'>, default=default_1)\n"
        "new: ConfigDynamicDefinition(keys=('path',), desc=path, of_type=<class 'str'>, default=default_2)"
    )
    with pytest.raises(ValueError, match=re.escape(msg)):
        config_set.add_config(keys="path", of_type=str, default="default_2", desc="path")


def test_config_dynamic_not_equal(conf_builder: ConfBuilder) -> None:
    config_set = conf_builder("")
    path = config_set.add_config(keys="path", of_type=Path, default=Path(), desc="path")
    paths = config_set.add_config(keys="paths", of_type=Path, default=Path(), desc="path")
    assert path != paths


def test_define_custom_set(tox_project: ToxProjectCreator) -> None:
    class MagicConfigSet(ConfigSet):
        SECTION = Section(None, "magic")

        def register_config(self) -> None:
            self.add_config("a", of_type=int, default=0, desc="number")
            self.add_config("b", of_type=str, default="", desc="string")

    project = tox_project({"tox.ini": "[testenv]\npackage=skip\n[A]\na=1\n[magic]\nb = ok"})
    result = project.run()
    section = MagicConfigSet.SECTION
    conf = result.state.conf.get_section_config(section, base=["A"], of_type=MagicConfigSet, for_env=None)
    assert conf["a"] == 1
    assert conf["b"] == "ok"
    exp = "MagicConfigSet(loaders=[IniLoader(section=magic, overrides={}), IniLoader(section=A, overrides={})])"
    assert repr(conf) == exp

    assert isinstance(result.state.conf._options, Parsed)  # noqa: SLF001


def test_do_not_allow_create_config_set(mocker: MockerFixture) -> None:
    with pytest.raises(TypeError, match="Can't instantiate"):
        ConfigSet(mocker.create_autospec(Config))  # ty: ignore[missing-argument] # intentionally wrong args to test TypeError


def test_set_env_raises_on_non_str(mocker: MockerFixture) -> None:
    env_set = EnvConfigSet(mocker.create_autospec(Config), Section("a", "b"), "b")
    env_set.loaders.insert(0, MemoryLoader(set_env=1))
    with pytest.raises(TypeError, match="1"):
        assert env_set["set_env"]


@pytest.mark.parametrize("work_dir", ["a", ""])
def test_config_work_dir(tox_project: ToxProjectCreator, work_dir: str) -> None:
    project = tox_project({"tox.ini": "[tox]\ntoxworkdir=b"})
    result = project.run("c", *(["--workdir", str(project.path / work_dir)] if work_dir else []))
    expected = project.path / work_dir if work_dir else Path("b")
    assert expected == result.state.conf.core["work_dir"]



from __future__ import annotations

import inspect
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar, cast

from tox.config.loader.api import ConfigLoadArgs, Loader, Override
from tox.config.loader.replacer import MatchError, replace
from tox.config.set_env import SetEnv
from tox.config.types import Command, EnvList
from tox.report import HandledError

from ._api import TomlTypes
from ._replace import TomlReplaceLoader, Unroll
from ._validate import validate

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from types import UnionType

    from tox.config.loader.convert import Factory
    from tox.config.loader.section import Section
    from tox.config.main import Config

_T = TypeVar("_T")
_V = TypeVar("_V")


class TomlLoader(Loader[TomlTypes]):
    """Load configuration from a pyproject.toml file."""

    def __init__(
        self,
        section: Section,
        overrides: list[Override],
        content: Mapping[str, TomlTypes],
        root_content: Mapping[str, TomlTypes],
        unused_exclude: set[str],
    ) -> None:
        self.content = content
        self._root_content = root_content
        self._unused_exclude = unused_exclude
        super().__init__(section, overrides)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.section.name}, {self.content!r})"

    def load_raw(self, key: str, conf: Config | None, env_name: str | None) -> TomlTypes:  # noqa: ARG002
        return self.content[key]

    def load_raw_from_root(self, path: str) -> TomlTypes:
        current = cast("TomlTypes", self._root_content)
        for key in path.split(self.section.SEP):
            if isinstance(current, dict):
                current = current[key]
            else:
                msg = f"Failed to load key {key} as not dictionary {current!r}"
                logging.warning(msg)
                raise KeyError(msg)
        return current

    def build(  # noqa: PLR0913
        self,
        key: str,
        of_type: type[_T] | UnionType,
        factory: Factory[_T],
        conf: Config | None,
        raw: TomlTypes,
        args: ConfigLoadArgs,
    ) -> _T:
        delay_replace = inspect.isclass(of_type) and issubclass(of_type, SetEnv)
        try:
            unroll = Unroll(conf=conf, loader=self, args=args)
            exploded = unroll(raw, skip_str=True) if delay_replace else unroll(raw)
            result = self.to(exploded, of_type, factory)
        except (HandledError, MatchError):
            raise
        except Exception as exception:
            name = "core" if args.env_name is None else args.env_name
            msg = f"failed to load {name}.{key}: {exception}"
            raise HandledError(msg) from exception
        if delay_replace:
            loader = self

            def _toml_replacer(value: str, args_: ConfigLoadArgs) -> str:
                if conf is None:
                    return value
                return replace(conf, TomlReplaceLoader(conf, loader), value, args_)

            result.use_replacer(_toml_replacer, args=args)
        return result

    def found_keys(self) -> set[str]:
        return set(self.content.keys()) - self._unused_exclude

    @staticmethod
    def to_str(value: TomlTypes) -> str:
        return validate(value, str)

    @staticmethod
    def to_bool(value: TomlTypes) -> bool:
        return validate(value, bool)

    @staticmethod
    def to_list(value: TomlTypes, of_type: type[_T]) -> Iterator[_T]:
        result = validate(value, cast("type[list[Any]]", list[of_type]))  # ty: ignore[invalid-type-form] # runtime generic from type variable
        return iter(cast("list[_T]", result))

    @staticmethod
    def to_set(value: TomlTypes, of_type: type[_T]) -> Iterator[_T]:
        result = validate(value, cast("type[list[Any]]", list[of_type]))  # ty: ignore[invalid-type-form] # runtime generic from type variable
        return iter(cast("list[_T]", result))

    @staticmethod
    def to_dict(value: TomlTypes, of_type: tuple[type[_T], type[_V]]) -> Iterator[tuple[_T, _V]]:
        result = validate(value, cast("type[dict[Any, Any]]", dict[of_type[0], of_type[1]]))  # ty: ignore[invalid-type-form] # runtime generic from type variables
        return iter(cast("dict[_T, _V]", result).items())

    @staticmethod
    def to_path(value: TomlTypes) -> Path:
        return Path(TomlLoader.to_str(value))

    @staticmethod
    def to_command(value: TomlTypes) -> Command | None:
        if value:
            return Command(args=cast("list[str]", value))  # validated during load in _ensure_type_correct
        return None

    @staticmethod
    def to_env_list(value: TomlTypes) -> EnvList:
        from ._product import expand_product  # noqa: PLC0415

        if not isinstance(value, list):
            msg = f"env_list must be a list, got {type(value).__name__}"
            raise TypeError(msg)
        envs: list[str] = []
        for item in value:
            if isinstance(item, str):
                envs.append(item)
            elif isinstance(item, dict) and "product" in item:
                envs.extend(expand_product(item))
            else:
                msg = f"env_list items must be strings or product dicts, got {type(item).__name__}"
                raise TypeError(msg)
        return EnvList(envs=envs)


__all__ = [
    "HandledError",
    "TomlLoader",
]



from __future__ import annotations

import os
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from tox.config.loader.api import Override
from tox.config.loader.memory import MemoryLoader
from tox.config.sets import ConfigSet
from tox.tox_env.python.pip.req_file import PythonDeps

if TYPE_CHECKING:
    from tests.conftest import ToxIniCreator
    from tox.config.main import Config
    from tox.pytest import ToxProjectCreator


def test_empty_config_repr(empty_config: Config) -> None:
    text = repr(empty_config)
    assert str(empty_config.core["tox_root"]) in text
    assert "config_source=ToxIni" in text


def test_empty_conf_tox_envs(empty_config: Config) -> None:
    tox_env_keys = list(empty_config)
    assert tox_env_keys == []


def test_empty_conf_get(empty_config: Config) -> None:
    result = empty_config.get_env("magic")
    assert isinstance(result, ConfigSet)
    loaders = result["base"]
    assert loaders == ["testenv"]


def test_config_some_envs(tox_ini_conf: ToxIniCreator) -> None:
    example = """
    [tox]
    env_list = py38, py37
    [testenv]
    deps = 1
        other: 2
    [testenv:magic]
    """
    config = tox_ini_conf(example)
    tox_env_keys = list(config)
    assert tox_env_keys == ["py38", "py37", "magic", "other"]

    config_set = config.get_env("py38")
    assert repr(config_set)
    assert isinstance(config_set, ConfigSet)
    assert list(config_set)


def test_config_overrides(tox_ini_conf: ToxIniCreator) -> None:
    conf = tox_ini_conf("[testenv]", override=[Override("testenv.c=ok")]).get_env("py")
    conf.add_config("c", of_type=str, default="d", desc="desc")
    assert conf["c"] == "ok"


def test_config_override_wins_memory_loader(tox_ini_conf: ToxIniCreator) -> None:
    main_conf = tox_ini_conf("[testenv]", override=[Override("testenv.c=ok")])
    conf = main_conf.get_env("py", loaders=[MemoryLoader(c="something_else")])
    conf.add_config("c", of_type=str, default="d", desc="desc")
    assert conf["c"] == "ok"


def test_config_override_appends_to_list(tox_ini_conf: ToxIniCreator) -> None:
    example = """
    [testenv]
    passenv = foo
    """
    conf = tox_ini_conf(example, override=[Override("testenv.passenv+=bar")]).get_env("testenv")
    conf.add_config("passenv", of_type=list[str], default=[], desc="desc")
    assert conf["passenv"] == ["foo", "bar"]


def test_config_override_sequence(tox_ini_conf: ToxIniCreator) -> None:
    example = """
    [testenv]
    passenv = foo
    """
    overrides = [Override("testenv.passenv=bar"), Override("testenv.passenv+=baz")]
    conf = tox_ini_conf(example, override=overrides).get_env("testenv")
    conf.add_config("passenv", of_type=list[str], default=[], desc="desc")
    assert conf["passenv"] == ["bar", "baz"]


def test_config_override_appends_to_empty_list(tox_ini_conf: ToxIniCreator) -> None:
    conf = tox_ini_conf("[testenv]", override=[Override("testenv.passenv+=bar")]).get_env("testenv")
    conf.add_config("passenv", of_type=list[str], default=[], desc="desc")
    assert conf["passenv"] == ["bar"]


@pytest.mark.parametrize(
    ("ini_key", "override_key"),
    [
        pytest.param("passenv", "pass_env", id="ini_old_override_new"),
        pytest.param("pass_env", "passenv", id="ini_new_override_old"),
    ],
)
def test_config_override_append_alias_key(tox_ini_conf: ToxIniCreator, ini_key: str, override_key: str) -> None:
    example = f"""
    [testenv]
    {ini_key} = foo
    """
    conf = tox_ini_conf(example, override=[Override(f"testenv.{override_key}+=bar")]).get_env("testenv")
    conf.add_config(["pass_env", "passenv"], of_type=list[str], default=[], desc="desc")
    result = conf["pass_env"]
    assert "foo" in result
    assert "bar" in result


def test_config_override_appends_to_setenv(tox_ini_conf: ToxIniCreator) -> None:
    example = """
    [testenv]
    setenv =
      foo = bar
    """
    conf = tox_ini_conf(example, override=[Override("testenv.setenv+=baz=quux")]).get_env("testenv")
    assert conf["setenv"].load("foo") == "bar"
    assert conf["setenv"].load("baz") == "quux"


def test_config_override_appends_to_setenv_multiple(tox_ini_conf: ToxIniCreator) -> None:
    example = """
    [testenv]
    setenv =
      foo = bar
    """
    overrides = [Override("testenv.setenv+=baz=quux"), Override("testenv.setenv+=less=more")]
    conf = tox_ini_conf(example, override=overrides).get_env("testenv")
    assert conf["setenv"].load("foo") == "bar"
    assert conf["setenv"].load("baz") == "quux"
    assert conf["setenv"].load("less") == "more"


def test_config_override_sequential_processing(tox_ini_conf: ToxIniCreator) -> None:
    example = """
    [testenv]
    setenv =
      foo = bar
    """
    overrides = [Override("testenv.setenv+=a=b"), Override("testenv.setenv=c=d"), Override("testenv.setenv+=e=f")]
    conf = tox_ini_conf(example, override=overrides).get_env("testenv")
    with pytest.raises(KeyError):
        assert conf["setenv"].load("foo") == "bar"
    with pytest.raises(KeyError):
        assert conf["setenv"].load("a") == "b"
    assert conf["setenv"].load("c") == "d"
    assert conf["setenv"].load("e") == "f"


def test_config_override_appends_to_empty_setenv(tox_ini_conf: ToxIniCreator) -> None:
    conf = tox_ini_conf("[testenv]", override=[Override("testenv.setenv+=foo=bar")]).get_env("testenv")
    assert conf["setenv"].load("foo") == "bar"


def test_config_override_appends_to_pythondeps(tox_ini_conf: ToxIniCreator, tmp_path: Path) -> None:
    example = """
    [testenv]
    deps = foo
    """
    conf = tox_ini_conf(example, override=[Override("testenv.deps+=bar")]).get_env("testenv")
    conf.add_config(
        "deps",
        of_type=PythonDeps,
        factory=partial(PythonDeps.factory, tmp_path),
        default=PythonDeps("", root=tmp_path),
        desc="desc",
    )
    assert conf["deps"].lines() == ["foo", "bar"]


def test_config_multiple_overrides(tox_ini_conf: ToxIniCreator, tmp_path: Path) -> None:
    example = """
    [testenv]
    deps = foo
    """
    overrides = [Override("testenv.deps+=bar"), Override("testenv.deps+=baz")]
    conf = tox_ini_conf(example, override=overrides).get_env("testenv")
    conf.add_config(
        "deps",
        of_type=PythonDeps,
        factory=partial(PythonDeps.factory, tmp_path),
        default=PythonDeps("", root=tmp_path),
        desc="desc",
    )
    assert conf["deps"].lines() == ["foo", "bar", "baz"]


def test_config_override_appends_to_empty_pythondeps(tox_ini_conf: ToxIniCreator, tmp_path: Path) -> None:
    example = """
    [testenv]
    """
    conf = tox_ini_conf(example, override=[Override("testenv.deps+=bar")]).get_env("testenv")
    conf.add_config(
        "deps",
        of_type=PythonDeps,
        factory=partial(PythonDeps.factory, tmp_path),
        default=PythonDeps("", root=tmp_path),
        desc="desc",
    )
    assert conf["deps"].lines() == ["bar"]


def test_config_override_cannot_append(tox_ini_conf: ToxIniCreator) -> None:
    example = """
    [testenv]
    foo = 1
    """
    conf = tox_ini_conf(example, override=[Override("testenv.foo+=2")]).get_env("testenv")
    conf.add_config("foo", of_type=int, default=0, desc="desc")
    with pytest.raises(ValueError, match="Only able to append to lists and dicts"):
        conf["foo"]


def test_args_are_paths_when_disabled(tox_project: ToxProjectCreator) -> None:
    ini = "[testenv]\npackage=skip\ncommands={posargs}\nargs_are_paths=False"
    project = tox_project({"tox.ini": ini, "w": {"a.txt": "a"}})
    args = "magic.py", str(project.path), f"..{os.sep}tox.ini", "..", f"..{os.sep}.."
    result = project.run("c", "-e", "py", "-k", "commands", "--", *args, from_cwd=project.path / "w")
    result.assert_success()
    assert result.out == f"[testenv:py]\ncommands = magic.py {project.path} ..{os.sep}tox.ini .. ..{os.sep}..\n"


def test_args_are_paths_when_from_child_dir(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv]\npackage=skip\ncommands={posargs}", "w": {"a.txt": "a"}})
    args = "magic.py", str(project.path), f"..{os.sep}tox.ini", "..", f"..{os.sep}.."
    result = project.run("c", "-e", "py", "-k", "commands", "--", *args, from_cwd=project.path / "w")
    result.assert_success()
    assert result.out == f"[testenv:py]\ncommands = magic.py {project.path} tox.ini . ..\n"


def test_args_are_paths_long_arg(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv]\npackage=skip\ncommands={posargs}", "w": {"a.txt": "a"}})
    long_arg = "a" * 300  # exceeds filesystem filename length limit (typically 255)
    result = project.run("c", "-e", "py", "-k", "commands", "--", long_arg, from_cwd=project.path / "w")
    result.assert_success()
    assert result.out == f"[testenv:py]\ncommands = {long_arg}\n"


def test_args_are_paths_when_with_change_dir(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv]\npackage=skip\ncommands={posargs}\nchange_dir=w", "w": {"a.txt": "a"}})
    args = "magic.py", str(project.path), "tox.ini", f"w{os.sep}a.txt", "w", "."
    result = project.run("c", "-e", "py", "-k", "commands", "--", *args)
    result.assert_success()
    assert result.out == f"[testenv:py]\ncommands = magic.py {project.path} ..{os.sep}tox.ini a.txt . ..\n"


@pytest.mark.timeout(60)
def test_get_env_reuses_config_across_package_flag(tox_project: ToxProjectCreator) -> None:
    project = tox_project({
        "tox.toml": """
env_list = ["a"]

[env.a]
package = "external"
package_env = ".build-a"

[env.".build-a"]
commands = [["python", "-c", "print('{env_tmp_dir}')"]]
package_glob = "{env_tmp_dir}/dist/*.whl"
""",
    })
    execute_calls = project.patch_execute(lambda r: 0)  # noqa: ARG005
    project.run("r", "-e", "a")
    calls = [(i[0][0].conf.name, i[0][3].cmd) for i in execute_calls.call_args_list]
    pkg_cmds = [cmd for name, cmd in calls if name == ".build-a"]
    assert pkg_cmds, f"no .build-a command found in {calls}"
    for cmd in pkg_cmds:
        cmd_str = " ".join(str(c) for c in cmd)
        assert "{env_tmp_dir}" not in cmd_str, f"env_tmp_dir not expanded in command: {cmd}"


def test_relative_config_paths_resolve(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[tox]"})
    ini = str(Path(project.path.name) / "tox.ini")
    result = project.run("c", "-c", ini, "-k", "change_dir", "env_dir", from_cwd=project.path.parent)
    result.assert_success()
    expected = f"[testenv:py]\nchange_dir = {project.path}\nenv_dir = {project.path / '.tox' / 'py'}\n"
    assert result.out == expected



from __future__ import annotations

import json
import sys
from typing import TYPE_CHECKING, Any

import pytest

from tox.session.cmd.show_config.json_format import colorize as colorize_json
from tox.session.cmd.show_config.toml_format import colorize as colorize_toml

if sys.version_info >= (3, 11):  # pragma: >=3.11 cover
    import tomllib
else:  # pragma: <3.11 cover
    import tomli as tomllib

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from tox.pytest import ToxProjectCreator

_FORMATS = [
    pytest.param("json", json.loads, id="json"),
    pytest.param("toml", tomllib.loads, id="toml"),
]


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_basic_structure(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true\n[testenv]\ncommands = python -c pass"})
    result = project.run("c", "-e", "py", "--format", fmt)
    result.assert_success()
    env = loader(result.out)["env"]["py"]
    assert env["type"] == "VirtualEnvRunner"
    assert env["env_name"] == "py"


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_core_with_flag(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true"})
    result = project.run("c", "-e", "py", "--core", "--format", fmt)
    result.assert_success()
    assert "tox_root" in loader(result.out)["tox"]


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_no_core_by_default(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true"})
    result = project.run("c", "-e", "py", "--format", fmt)
    result.assert_success()
    assert "tox" not in loader(result.out)


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_core_with_all_envs(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true\nenv_list = py"})
    result = project.run("c", "-e", "ALL", "--format", fmt)
    result.assert_success()
    assert "tox" in loader(result.out)


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_key_filtering(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true\n[testenv]\ncommands = python -c pass"})
    result = project.run("c", "-e", "py", "--format", fmt, "-k", "env_name", "commands")
    result.assert_success()
    env = loader(result.out)["env"]["py"]
    assert "env_name" in env
    assert "commands" in env
    assert "type" not in env


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_native_types(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true\n[testenv]\npackage = skip"})
    result = project.run("c", "-e", "py", "--format", fmt, "-k", "suicide_timeout", "allowlist_externals")
    result.assert_success()
    env = loader(result.out)["env"]["py"]
    assert isinstance(env["suicide_timeout"], float)
    assert isinstance(env["allowlist_externals"], list)


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_multiple_envs(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true\nenv_list = a,b\n[testenv]\npackage = skip"})
    result = project.run("c", "-e", "a,b", "--format", fmt)
    result.assert_success()
    data = loader(result.out)
    assert "a" in data["env"]
    assert "b" in data["env"]


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_unused_keys(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[testenv:py]\nmagical = yes\nmagic = yes"})
    result = project.run("c", "-e", "py", "--format", fmt)
    result.assert_success()
    assert sorted(loader(result.out)["env"]["py"]["unused"]) == ["magic", "magical"]


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_no_unused_with_key_filter(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[testenv:py]\nmagical = yes"})
    result = project.run("c", "-e", "py", "--format", fmt, "-k", "env_name")
    result.assert_success()
    assert "unused" not in loader(result.out)["env"]["py"]


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_output_to_file(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any], tmp_path: Path) -> None:
    out_file = tmp_path / f"config.{fmt}"
    project = tox_project({"tox.ini": "[tox]\nno_package = true"})
    result = project.run("c", "-e", "py", "--format", fmt, "-o", str(out_file))
    result.assert_success()
    assert not result.out
    assert "py" in loader(out_file.read_text())["env"]


@pytest.mark.parametrize("fmt", [pytest.param("json", id="json"), pytest.param("toml", id="toml")])
def test_file_output_no_ansi(tox_project: ToxProjectCreator, fmt: str, tmp_path: Path) -> None:
    out_file = tmp_path / f"config.{fmt}"
    project = tox_project({"tox.ini": "[tox]\nno_package = true"})
    result = project.run("c", "-e", "py", "--format", fmt, "-o", str(out_file))
    result.assert_success()
    assert "\x1b[" not in out_file.read_text()


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_set_env(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true\n[testenv]\nset_env =\n    A=1\n    B=hello"})
    result = project.run("c", "-e", "py", "--format", fmt, "-k", "set_env")
    result.assert_success()
    se = loader(result.out)["env"]["py"]["set_env"]
    assert se["A"] == "1"
    assert se["B"] == "hello"


@pytest.mark.parametrize(("fmt", "loader"), _FORMATS)
def test_deps(tox_project: ToxProjectCreator, fmt: str, loader: Callable[[str], Any]) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true\n[testenv]\ndeps =\n    pytest\n    flask>=2.0"})
    result = project.run("c", "-e", "py", "--format", fmt, "-k", "deps")
    result.assert_success()
    deps = loader(result.out)["env"]["py"]["deps"]
    assert isinstance(deps, list)
    assert "pytest" in deps
    assert "flask>=2.0" in deps


def test_json_exception(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv:a]\nbase_python = missing-python"})
    result = project.run("c", "-e", "a", "--format", "json", "-k", "env_site_packages_dir", raise_on_config_fail=False)
    result.assert_failed(code=-1)
    assert "error" in json.loads(result.out)["env"]["a"]["env_site_packages_dir"]


def test_json_pass_env(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true\n[testenv]\npass_env = HOME,PATH"})
    result = project.run("c", "-e", "py", "--format", "json", "-k", "pass_env")
    result.assert_success()
    pe = json.loads(result.out)["env"]["py"]["pass_env"]
    assert isinstance(pe, list)
    assert "HOME" in pe


def test_json_alias_key(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true"})
    result = project.run("c", "-e", "py", "--format", "json", "-k", "setenv")
    result.assert_success()
    assert "set_env" in json.loads(result.out)["env"]["py"]


def test_json_valid_output(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true"})
    result = project.run("c", "-e", "py", "--format", "json")
    result.assert_success()
    json.loads(result.out)


def test_json_native_types_approx(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true\n[testenv]\npackage = skip"})
    result = project.run("c", "-e", "py", "--format", "json", "-k", "suicide_timeout")
    result.assert_success()


@pytest.mark.parametrize("fmt", [pytest.param("json", id="json"), pytest.param("toml", id="toml")])
def test_key_filter_skips_missing(tox_project: ToxProjectCreator, fmt: str) -> None:
    project = tox_project({"tox.ini": "[tox]\nno_package = true"})
    result = project.run("c", "-e", "py", "--format", fmt, "-k", "no_such_key")
    result.assert_success()


def test_core_exception_json(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": "[testenv:a]\nbase_python = missing-python"})
    result = project.run(
        "c", "-e", "a", "--core", "--format", "json", "-k", "env_site_packages_dir", raise_on_config_fail=False
    )
    result.assert_failed(code=-1)
    data = json.loads(result.out)
    assert "error" in data["env"]["a"]["env_site_packages_dir"]
    assert "tox" in data


def testcolorize_json() -> None:
    result = colorize_json('{\n  "key": "val"\n}')
    assert "\x1b[" in result
    assert "key" in result


def testcolorize_toml() -> None:
    result = colorize_toml("[env.py]\nname = true")
    assert "\x1b[" in result
    assert "env.py" in result



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import pytest

from tox.execute.request import ExecuteRequest, StdinSource

if TYPE_CHECKING:
    from unittest.mock import MagicMock

    from tox.pytest import ToxProjectCreator


@pytest.mark.parametrize("trail", [[], ["--"]], ids=["no_posargs", "empty_posargs"])
def test_exec_fail_no_posargs(tox_project: ToxProjectCreator, trail: list[str]) -> None:
    outcome = tox_project({"tox.ini": ""}).run("e", "-e", "py39", *trail)
    outcome.assert_failed()
    msg = "ROOT: HandledError| You must specify a command as positional arguments, use -- <command>\n"
    outcome.assert_out_err(msg, "")


def test_exec_fail_multiple_target(tox_project: ToxProjectCreator) -> None:
    outcome = tox_project({"tox.ini": ""}).run("e", "-e", "py39,py38", "--", "py")
    outcome.assert_failed()
    msg = "ROOT: HandledError| exactly one target environment allowed in exec mode but found py39, py38\n"
    outcome.assert_out_err(msg, "")


@pytest.mark.parametrize("exit_code", [1, 0])
def test_exec(tox_project: ToxProjectCreator, exit_code: int) -> None:
    prj = tox_project({"tox.ini": "[testenv]\npackage=skip"})
    py_cmd = f"import sys; print(sys.version); raise SystemExit({exit_code})"
    outcome = prj.run("e", "-e", "py", "--", "python", "-c", py_cmd)
    if exit_code:
        outcome.assert_failed()
    else:
        outcome.assert_success()
    assert sys.version in outcome.out


def test_exec_help(tox_project: ToxProjectCreator) -> None:
    outcome = tox_project({"tox.ini": ""}).run("e", "-h")
    outcome.assert_success()


def test_exec_always_no_capture(tox_project: ToxProjectCreator) -> None:
    """Verify tox exec always runs with no_capture enabled for interactive mode."""
    ini = "[testenv]\npackage=skip"
    project = tox_project({"tox.ini": ini})

    captured_options: list[MagicMock] = []

    def capture_options(request):  # noqa: ANN001, ANN202
        captured_options.append(request)
        return 0

    execute_calls = project.patch_execute(capture_options)
    result = project.run("e", "-e", "py", "--", "python", "--version")
    result.assert_success()

    assert execute_calls.call_count > 0
    for call in execute_calls.call_args_list:
        _, kwargs = call
        env_instance = kwargs.get("self")
        if env_instance and hasattr(env_instance, "options"):
            assert env_instance.options.no_capture is True


def test_exec_passes_stdin_through(tox_project: ToxProjectCreator, monkeypatch: pytest.MonkeyPatch) -> None:
    project = tox_project({"tox.ini": "[testenv]\npackage=skip"})
    monkeypatch.setattr(sys.stdin, "isatty", lambda: False)

    stdin_sources: list[StdinSource] = []

    def capture_stdin(request: ExecuteRequest) -> int:
        stdin_sources.append(request.stdin)
        return 0

    project.patch_execute(capture_stdin)
    result = project.run("e", "-e", "py", "--", "python", "--version")
    result.assert_success()

    assert stdin_sources
    assert all(source == StdinSource.USER for source in stdin_sources)



from __future__ import annotations

import os
import sys
from itertools import chain, combinations
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from tox.pytest import MonkeyPatch, ToxProjectCreator, check_os_environ
from tox.report import HandledError

if TYPE_CHECKING:
    from collections.abc import Sequence
    from pathlib import Path

    from pytest_mock import MockerFixture


def test_tox_project_no_base(tox_project: ToxProjectCreator) -> None:
    project = tox_project(
        {
            "tox.ini": "[tox]",
            "src": {"__init__.py": "pass", "a": "out", "b": {"c": "out"}, "e": {"f": ""}},
        },
    )
    assert str(project.path) in repr(project)
    assert project.path.exists()
    assert project.structure == {
        "tox.ini": "[tox]",
        "src": {"__init__.py": "pass", "a": "out", "e": {"f": ""}, "b": {"c": "out"}},
    }


def test_tox_project_base(tmp_path: Path, tox_project: ToxProjectCreator) -> None:
    base = tmp_path / "base"
    base.mkdir()
    (base / "out").write_text("a")
    project = tox_project({"tox.ini": "[tox]"}, base=base)
    assert project.structure


COMB = list(chain.from_iterable(combinations(["DIFF", "MISS", "EXTRA"], i) for i in range(4)))


@pytest.mark.parametrize("ops", COMB, ids=["-".join(i) for i in COMB])
def test_env_var(monkeypatch: MonkeyPatch, ops: list[str]) -> None:
    with monkeypatch.context() as m:
        if "DIFF" in ops:
            m.setenv("DIFF", "B")
        if "MISS" in ops:
            m.setenv("MISS", "1")
        m.setenv("NO_CHANGE", "yes")
        m.setenv("PYTHONPATH", "yes")  # values to clean before run

        with check_os_environ():
            assert "PYTHONPATH" not in os.environ
            if "EXTRA" in ops:
                m.setenv("EXTRA", "A")
            if "DIFF" in ops:
                m.setenv("DIFF", "D")
            if "MISS" in ops:
                m.delenv("MISS")

            from tox.pytest import pytest as tox_pytest  # noqa: PLC0415

            exp = "test changed environ"
            if "EXTRA" in ops:
                exp += " extra {'EXTRA': 'A'}"
            if "MISS" in ops:
                exp += " miss {'MISS': '1'}"
            if "DIFF" in ops:
                exp += " diff {'DIFF = B vs D'}"

            def fail(msg: str) -> None:
                assert msg == exp

            m.setattr(tox_pytest, "fail", fail)
        assert "PYTHONPATH" in os.environ


def test_tox_run_does_not_return_exit_code(tox_project: ToxProjectCreator, mocker: MockerFixture) -> None:
    project = tox_project({"tox.ini": ""})
    mocker.patch("tox.run.main", return_value=None)
    with pytest.raises(RuntimeError, match="exit code not set"):
        project.run("c")


def test_tox_run_fails_before_state_setup(tox_project: ToxProjectCreator, mocker: MockerFixture) -> None:
    project = tox_project({"tox.ini": ""})
    mocker.patch("tox.run.main", side_effect=HandledError("something went bad"))
    outcome = project.run("c")
    with pytest.raises(RuntimeError, match="no state"):
        assert outcome.state


def test_tox_run_outcome_repr(tox_project: ToxProjectCreator) -> None:
    project = tox_project({"tox.ini": ""})
    outcome = project.run("l")
    msg = dedent(
        f"""
    code: 0
    cmd: {sys.executable} -m tox l
    cwd: {project.path}
    standard output
    default environments:
    py -> [no description]
    """,
    ).lstrip()
    assert repr(outcome) == msg
    assert outcome.shell_cmd == f"{sys.executable} -m tox l"


def test_tox_run_assert_out_err_no_dedent(tox_project: ToxProjectCreator, mocker: MockerFixture) -> None:
    project = tox_project({"tox.ini": ""})

    def _main(args: Sequence[str]) -> int:  # noqa: ARG001
        print(" goes on out", file=sys.stdout)  # noqa: T201
        print(" goes on err", file=sys.stderr)  # noqa: T201
        return 0

    mocker.patch("tox.run.main", side_effect=_main)
    outcome = project.run("c")
    outcome.assert_out_err(" goes on out\n", " goes on err\n", dedent=False)



"""Record information about tox environments."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from tox.execute import Outcome


class EnvJournal:
    """Report the status of a tox environment."""

    def __init__(self, enabled: bool, name: str) -> None:  # noqa: FBT001
        self._enabled = enabled
        self.name = name
        self._content: dict[str, Any] = {}
        self._executes: list[tuple[str, Outcome]] = []

    def __setitem__(self, key: str, value: Any) -> None:
        """Add a new entry under key into the event journal.

        :param key: the key under what to add the data
        :param value: the data to add

        """
        self._content[key] = value

    def __bool__(self) -> bool:
        """:returns: a flag indicating if the event journal is on or not"""
        return self._enabled

    def add_execute(self, outcome: Outcome, run_id: str) -> None:
        """Add a command execution to the journal.

        :param outcome: the execution outcome
        :param run_id: the execution id

        """
        self._executes.append((run_id, outcome))

    @property
    def content(self) -> dict[str, Any]:
        """:returns: the env journal content (merges explicit keys and execution commands)"""
        tests: list[dict[str, Any]] = []
        setup: list[dict[str, Any]] = []
        for run_id, outcome in self._executes:
            one = {
                "command": outcome.cmd,
                "output": outcome.out,
                "err": outcome.err,
                "retcode": outcome.exit_code,
                "elapsed": outcome.elapsed,
                "show_on_standard": outcome.show_on_standard,
                "run_id": run_id,
                "start": outcome.start,
                "end": outcome.end,
            }
            if run_id.startswith(("commands", "build")):
                tests.append(one)
            else:
                setup.append(one)
        if tests:
            self["test"] = tests
        if setup:
            self["setup"] = setup
        return self._content


__all__ = ("EnvJournal",)



from __future__ import annotations

import locale
import logging
import os
import signal
import sys
from io import TextIOWrapper
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import MagicMock

from tox.execute.local_sub_process import LocalSubProcessExecutor
from tox.execute.request import ExecuteRequest, StdinSource
from tox.report import NamedBytesIO

if TYPE_CHECKING:
    from types import FrameType

    from tox.execute import Outcome

logging.basicConfig(level=logging.DEBUG, format="%(relativeCreated)d\t%(levelname).1s\t%(message)s")
bad_process = Path(__file__).parent / "bad_process.py"

executor = LocalSubProcessExecutor(colored=False)
request = ExecuteRequest(
    cmd=[sys.executable, bad_process, sys.argv[1]],
    cwd=Path().absolute(),
    env=os.environ.copy(),
    stdin=StdinSource.API,
    run_id="",
)
out_err = (
    TextIOWrapper(NamedBytesIO("out"), encoding=locale.getpreferredencoding(False)),
    TextIOWrapper(NamedBytesIO("err"), encoding=locale.getpreferredencoding(False)),
)


def show_outcome(outcome: Outcome | None) -> None:
    if outcome is not None:  # pragma: no branch
        print(outcome.exit_code)  # noqa: T201
        print(repr(outcome.out))  # noqa: T201
        print(repr(outcome.err))  # noqa: T201
        print(outcome.elapsed, end="")  # noqa: T201
        print("done show outcome", file=sys.stderr)  # noqa: T201


def handler(s: int, f: FrameType | None) -> None:
    logging.info("signal %s at %s", s, f)
    global interrupt_done  # noqa: PLW0603
    if interrupt_done is False:  # pragma: no branch
        interrupt_done = True
        logging.info("interrupt via %s", status)
        status.interrupt()
        logging.info("interrupt finished via %s", status)


interrupt_done = False
signal.signal(signal.SIGINT, handler)
logging.info("PID %d start %r", os.getpid(), request)
tox_env = MagicMock(conf={"suicide_timeout": 0.01, "interrupt_timeout": 0.05, "terminate_timeout": 0.07})
tox_env.options.no_capture = False
try:
    with executor.call(request, show=False, out_err=out_err, env=tox_env) as status:
        logging.info("wait on %r", status)
        while status.exit_code is None:
            status.wait(timeout=0.01)  # use wait here with timeout to not block the main thread
        logging.info("wait over on %r", status)
    show_outcome(status.outcome)
except Exception as exception:  # pragma: no cover
    logging.exception(exception)  # noqa: TRY401 # pragma: no cover
finally:
    logging.info("done")
    logging.shutdown()



"""A tox python environment runner that uses the virtualenv project."""

from __future__ import annotations

from typing import TYPE_CHECKING

from tox.plugin import impl
from tox.tox_env.python.runner import PythonRun

from .api import VirtualEnv

if TYPE_CHECKING:
    from pathlib import Path

    from tox.tox_env.register import ToxEnvRegister


class VirtualEnvRunner(VirtualEnv, PythonRun):
    """local file system python virtual environment via the virtualenv package."""

    @staticmethod
    def id() -> str:
        return "virtualenv"

    @property
    def _package_tox_env_type(self) -> str:
        return "virtualenv-pep-517"

    @property
    def _external_pkg_tox_env_type(self) -> str:
        return "virtualenv-cmd-builder"

    @property
    def default_pkg_type(self) -> str:
        tox_root: Path = self.core["tox_root"]
        if not (any((tox_root / i).exists() for i in ("pyproject.toml", "setup.py", "setup.cfg"))):
            return "skip"
        return super().default_pkg_type


@impl
def tox_register_tox_env(register: ToxEnvRegister) -> None:
    register.add_run_env(VirtualEnvRunner)



"""A tox run environment that handles the Python language."""

from __future__ import annotations

import logging
from abc import ABC
from functools import partial
from typing import TYPE_CHECKING

from packaging.utils import canonicalize_name

from tox.config.loader.str_convert import StrConvert
from tox.config.types import Command
from tox.execute import Outcome
from tox.report import HandledError
from tox.session.cmd.run.single import run_command_set
from tox.tox_env.errors import Fail, Skip
from tox.tox_env.python.pip.req_file import PythonDeps
from tox.tox_env.python.pylock import Pylock
from tox.tox_env.runner import RunToxEnv

from .api import Python
from .dependency_groups import resolve as resolve_dependency_groups
from .extras import resolve_extras_static

if TYPE_CHECKING:
    from pathlib import Path

    from tox.config.cli.parser import Parsed
    from tox.config.main import Config
    from tox.config.sets import CoreConfigSet, EnvConfigSet
    from tox.tox_env.api import ToxEnvCreateArgs
    from tox.tox_env.package import Package


class PythonRun(Python, RunToxEnv, ABC):
    def __init__(self, create_args: ToxEnvCreateArgs) -> None:
        super().__init__(create_args)

    def register_config(self) -> None:
        super().register_config()
        root = self.core["toxinidir"]
        self.conf.add_config(
            keys=["deps"],
            of_type=PythonDeps,
            factory=partial(PythonDeps.factory, root),
            default=PythonDeps("", root),
            desc="python dependencies with optional version specifiers, as specified by PEP-440",
        )
        self.conf.add_config(
            keys=["dependency_groups"],
            of_type=set[str],
            default=set(),
            desc="dependency groups to install of the target package",
            post_process=_normalize_extras,
        )
        self.conf.add_config(
            keys=["extras"],
            of_type=set[str],
            default=set(),
            desc="extras to install of the target package",
            post_process=_normalize_extras,
        )

        def _validate_pylock_not_with_deps(value: str) -> str:
            if value and self.conf["deps"].lines():
                msg = "cannot use both 'deps' and 'pylock' in the same environment"
                raise Fail(msg)
            return value

        self.conf.add_config(
            keys=["pylock"],
            of_type=str,
            default="",
            desc="PEP 751 pylock.toml lock file path to install locked dependencies from",
            post_process=_validate_pylock_not_with_deps,
        )
        self.conf.add_config(
            keys=["extra_setup_commands"],
            of_type=list[Command],
            default=[],
            desc="commands to execute after setup (deps and package install) but before test commands",
        )
        add_skip_missing_interpreters_to_core(self.core, self.options)
        add_skip_missing_interpreters_to_env(self.conf, self.core, self.options)

    @property
    def _package_types(self) -> tuple[str, ...]:
        return "wheel", "sdist", "sdist-wheel", "editable", "editable-legacy", "deps-only", "skip", "external"

    def _register_package_conf(self) -> bool:
        # provision package type
        desc = f"package installation mode - {' | '.join(i for i in self._package_types)} "
        if not super()._register_package_conf():
            self.conf.add_constant(["package"], desc, "skip")
            return False
        if getattr(self.options, "install_pkg", None) is not None:
            self.conf.add_constant(["package"], desc, "external")
        else:
            self.conf.add_config(
                keys=["use_develop", "usedevelop"],
                desc="use develop mode",
                default=False,
                of_type=bool,
            )
            develop_mode = self.conf["use_develop"] or getattr(self.options, "develop", False)
            if develop_mode:
                self.conf.add_constant(["package"], desc, "editable")
            else:
                self.conf.add_config(keys="package", of_type=str, default=self.default_pkg_type, desc=desc)

        return self.pkg_type != "skip"

    @property
    def default_pkg_type(self) -> str:
        return "sdist"

    @property
    def pkg_type(self) -> str:
        pkg_type: str = self.conf["package"]
        if pkg_type not in self._package_types:
            values = ", ".join(self._package_types)
            msg = f"invalid package config type {pkg_type} requested, must be one of {values}"
            raise HandledError(msg)
        return pkg_type

    def _setup_pkg(self) -> None:
        if self.pkg_type == "deps-only":
            self._install_package_deps_only()
            return
        super()._setup_pkg()

    def _install_package_deps_only(self) -> None:
        extras: set[str] = self.conf["extras"]
        root: Path = self.core["package_root"]
        if (deps := resolve_extras_static(root, extras)) is None:
            package_env = self.package_env
            assert package_env is not None  # noqa: S101
            with package_env.display_context(self._has_display_suspended):
                deps = package_env.load_deps_for_env(self.conf)
        if deps and not self.options.package_only:
            self._install(deps, PythonRun.__name__, "package_deps")

    def _setup_env(self) -> None:
        super()._setup_env()
        if getattr(self.options, "skip_env_install", False):
            logging.warning("skip installing dependencies and package")
            return
        if self.conf["pylock"]:
            self._install_pylock()
        else:
            self._install_deps()
            self._install_dependency_groups()

    def _install_deps(self) -> None:
        requirements_file: PythonDeps = self.conf["deps"]
        self._install(requirements_file, PythonRun.__name__, "deps")

    def _install_dependency_groups(self) -> None:
        groups: set[str] = self.conf["dependency_groups"]
        if not groups:
            return
        try:
            root: Path = self.core["package_root"]
        except KeyError:
            root = self.core["tox_root"]
        requirements = resolve_dependency_groups(root, groups)
        self._install(list(requirements), PythonRun.__name__, "dependency-groups")

    def _install_pylock(self) -> None:
        pylock_path: str = self.conf["pylock"]
        try:
            root: Path = self.core["package_root"]
        except KeyError:
            root = self.core["tox_root"]
        if not (path := root / pylock_path).exists():
            msg = f"pylock file {pylock_path!r} not found at {path}"
            raise Fail(msg)
        info = self.base_python
        marker_env = {
            "implementation_name": info.impl_lower,
            "platform_python_implementation": info.implementation,
            "python_version": info.version_dot,
            "python_full_version": f"{info.version_info.major}.{info.version_info.minor}.{info.version_info.micro}",
            "sys_platform": info.platform,
        }
        extras: set[str] = self.conf["extras"]
        groups: set[str] = self.conf["dependency_groups"]
        pylock = Pylock(path=path, extras=frozenset(extras), groups=frozenset(groups), marker_env=marker_env)
        self._install(pylock, PythonRun.__name__, "pylock")

    def _setup_with_env(self) -> None:
        super()._setup_with_env()
        self._run_extra_setup_commands()

    def _run_extra_setup_commands(self) -> None:
        command_set: list[Command] = self.conf["extra_setup_commands"]
        if not command_set:
            return
        chdir: Path = self.conf["change_dir"]
        chdir.mkdir(exist_ok=True, parents=True)
        ignore_errors: bool = self.conf["ignore_errors"]
        outcomes: list[Outcome] = []
        exit_code = run_command_set(self, "extra_setup_commands", chdir, ignore_errors, outcomes)
        if exit_code != Outcome.OK and not ignore_errors:
            msg = "extra_setup_commands failed"
            raise Fail(msg)

    def _build_packages(self) -> list[Package]:
        package_env = self.package_env
        assert package_env is not None  # noqa: S101
        with package_env.display_context(self._has_display_suspended):
            try:
                packages = package_env.perform_packaging(self.conf)
            except Skip as exception:
                msg = f"{exception.args[0]} for package environment {package_env.conf['env_name']}"
                raise Skip(msg) from exception
        return packages


def add_skip_missing_interpreters_to_core(core: CoreConfigSet, options: Parsed) -> None:
    def skip_missing_interpreters_post_process(value: bool) -> bool:  # noqa: FBT001
        if getattr(options, "skip_missing_interpreters", "config") != "config":
            return StrConvert().to_bool(options.skip_missing_interpreters)
        return value

    core.add_config(
        keys=["skip_missing_interpreters"],
        default=True,
        of_type=bool,
        post_process=skip_missing_interpreters_post_process,
        desc="skip running missing interpreters",
    )


def add_skip_missing_interpreters_to_env(conf: EnvConfigSet, core: CoreConfigSet, options: Parsed) -> None:
    def _default_skip_missing(conf: Config, env_name: str | None) -> bool:  # noqa: ARG001
        return core["skip_missing_interpreters"]

    def _post_process(value: bool) -> bool:  # noqa: FBT001
        if getattr(options, "skip_missing_interpreters", "config") != "config":
            return StrConvert().to_bool(options.skip_missing_interpreters)
        return value

    conf.add_config(  # ty: ignore[no-matching-overload] # https://github.com/astral-sh/ty/issues/2428
        keys=["skip_missing_interpreters"],
        default=_default_skip_missing,
        of_type=bool,
        post_process=_post_process,
        desc="override core skip_missing_interpreters for this environment",
    )


def add_extras_to_env(conf: EnvConfigSet) -> None:
    conf.add_config(
        keys=["extras"],
        of_type=set[str],
        default=set(),
        desc="extras to install of the target package",
        post_process=_normalize_extras,
    )


def _normalize_extras(values: set[str]) -> set[str]:
    # although _ and . is allowed this will be normalized during packaging to -
    # https://packaging.python.org/en/latest/specifications/dependency-specifiers/#grammar
    return {canonicalize_name(v) for v in values}


__all__ = [
    "PythonRun",
    "add_extras_to_env",
    "add_skip_missing_interpreters_to_core",
    "add_skip_missing_interpreters_to_env",
]



"""Provides configuration values from the environment variables."""

from __future__ import annotations

import logging
import os
from typing import Any

from tox.config.loader.str_convert import StrConvert

CONVERT = StrConvert()


def get_env_var(key: str, of_type: type[Any]) -> tuple[Any, str] | None:
    """Get the environment variable option.

    :param key: the config key requested
    :param of_type: the type we would like to convert it to

    :returns: the converted value and its origin, or ``None`` if not found

    """
    key_upper = key.upper()
    for environ_key in (f"TOX_{key_upper}", f"TOX{key_upper}"):
        if environ_key in os.environ:
            value = os.environ[environ_key]
            origin = getattr(of_type, "__origin__", of_type.__class__)
            try:
                if origin is list:
                    entry_type = of_type.__args__[0]
                    result = [CONVERT.to(raw=v, of_type=entry_type, factory=None) for v in value.split(";")]
                else:
                    result = CONVERT.to(raw=value, of_type=of_type, factory=None)
            except Exception as exception:  # noqa: BLE001
                logging.warning(
                    "env var %s=%r cannot be transformed to %r because %r",
                    environ_key,
                    value,
                    of_type,
                    exception,
                )
            else:
                return result, f"env var {environ_key}"
    return None


__all__ = ("get_env_var",)



"""Show materialized configuration of tox environments."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Literal, get_args

from tox.plugin import impl
from tox.session.cmd.run.common import env_run_create_flags
from tox.session.env_select import CliEnv, register_env_select_flags

if TYPE_CHECKING:
    from tox.config.cli.parser import ToxParser
    from tox.session.state import State

ConfigFormat = Literal["ini", "json", "toml"]


@impl
def tox_add_option(parser: ToxParser) -> None:
    our = parser.add_command("config", ["c"], "show tox configuration", show_config)
    our.add_argument(
        "-k",
        nargs="+",
        help="list just configuration keys specified",
        dest="list_keys_only",
        default=[],
        metavar="key",
    )
    our.add_argument(
        "--core",
        action="store_true",
        help="show core options (by default is hidden unless -e ALL is passed)",
        dest="show_core",
    )
    our.add_argument(
        "--format",
        choices=get_args(ConfigFormat),
        default="ini",
        help="output format (default: %(default)s)",
        dest="config_format",
    )
    our.add_argument(
        "-o",
        "--output-file",
        of_type=Path,
        default=None,
        help="write output to file instead of stdout",
        dest="output_file",
    )
    register_env_select_flags(our, default=CliEnv())
    env_run_create_flags(our, mode="config")


def show_config(state: State) -> int:
    from .ini import show_config_ini  # noqa: PLC0415
    from .json_format import show_config_json  # noqa: PLC0415
    from .toml_format import show_config_toml  # noqa: PLC0415

    fmt: ConfigFormat = state.conf.options.config_format
    if fmt == "json":
        return show_config_json(state)
    if fmt == "toml":
        return show_config_toml(state)
    return show_config_ini(state)



"""On UNIX we use selectors to drain streams efficiently, following CPython's subprocess implementation."""

from __future__ import annotations

import contextlib  # pragma: win32 no cover
import errno  # pragma: win32 no cover
import os  # pragma: win32 no cover
import selectors  # pragma: win32 no cover
from typing import TYPE_CHECKING, Any

from .read_via_thread import ReadViaThread  # pragma: win32 no cover

if TYPE_CHECKING:
    from collections.abc import Callable

TIMEOUT_FOR_INTERRUPT = 0.05  # pragma: win32 no cover
READ_CHUNK_SIZE = 32768  # pragma: win32 no cover


class ReadViaThreadUnix(ReadViaThread):  # pragma: win32 no cover
    def __init__(self, file_no: int, handler: Callable[[bytes], int], name: str, drain: bool) -> None:  # noqa: FBT001
        super().__init__(file_no, handler, name, drain)

    def _read_stream(self) -> None:
        selector = selectors.DefaultSelector()
        try:
            selector.register(self.file_no, selectors.EVENT_READ)
        except (OSError, ValueError):  # pragma: no cover
            return

        try:
            self._read_until_eof(selector)
        finally:
            selector.close()

    def _read_until_eof(self, selector: selectors.DefaultSelector) -> None:
        while selector.get_map() and not self.stop.is_set():
            try:
                ready = selector.select(timeout=TIMEOUT_FOR_INTERRUPT)
            except (InterruptedError, OSError) as exception:
                if isinstance(exception, OSError) and exception.errno != errno.EINTR:
                    raise
                continue

            if not ready:
                continue

            for key, _ in ready:
                self._read_chunk(selector, key)

    def _drain_stream(self) -> None:
        selector = selectors.DefaultSelector()
        try:
            selector.register(self.file_no, selectors.EVENT_READ)
        except (OSError, ValueError):  # pragma: no cover
            return

        try:
            while selector.get_map():
                try:
                    ready = selector.select(timeout=0)
                except (InterruptedError, OSError) as exception:
                    if isinstance(exception, OSError) and exception.errno != errno.EINTR:  # pragma: no cover
                        raise  # pragma: no cover
                    continue

                if not ready:
                    break

                for key, _ in ready:
                    self._read_chunk(selector, key)
        finally:
            selector.close()

    def _read_chunk(self, selector: selectors.DefaultSelector, key: selectors.SelectorKey) -> None:
        try:
            data = os.read(key.fd, READ_CHUNK_SIZE)
        except OSError as exception:
            if exception.errno == errno.EINTR:
                return
            if exception.errno not in {errno.EBADF, errno.EIO}:  # pragma: no cover
                raise  # pragma: no cover
            data = b""

        if data:
            self.handler(data)
        else:
            self._safe_unregister(selector, key.fileobj)

    @staticmethod
    def _safe_unregister(selector: selectors.DefaultSelector, fileobj: Any) -> None:
        with contextlib.suppress(KeyError, ValueError):
            selector.unregister(fileobj)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from tox.pytest import ToxProjectCreator


def test_recreate_commands_run_on_recreate(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """
[env_run_base]
package = "skip"
recreate_commands = [["python", "-c", "print('RECREATE_CLEANUP')"]]
commands = [["python", "-c", "print('MAIN')"]]
""",
    })
    first = proj.run("r")
    first.assert_success()
    assert "RECREATE_CLEANUP" not in first.out

    second = proj.run("r", "-r")
    second.assert_success()
    assert "RECREATE_CLEANUP" in second.out
    assert "MAIN" in second.out


def test_recreate_commands_not_run_on_first_creation(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """
[env_run_base]
package = "skip"
recreate_commands = [["python", "-c", "print('RECREATE_CLEANUP')"]]
""",
    })
    result = proj.run("r")
    result.assert_success()
    assert "RECREATE_CLEANUP" not in result.out


def test_recreate_commands_failure_does_not_block_recreation(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """
[env_run_base]
package = "skip"
recreate_commands = [["python", "-c", "raise SystemExit(1)"]]
commands = [["python", "-c", "print('MAIN')"]]
""",
    })
    first = proj.run("r")
    first.assert_success()

    second = proj.run("r", "-r")
    second.assert_success()
    assert "MAIN" in second.out


def test_recreate_commands_not_run_without_recreate(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """
[env_run_base]
package = "skip"
recreate_commands = [["python", "-c", "print('RECREATE_CLEANUP')"]]
commands = [["python", "-c", "print('MAIN')"]]
""",
    })
    first = proj.run("r")
    first.assert_success()

    second = proj.run("r")
    second.assert_success()
    assert "RECREATE_CLEANUP" not in second.out
    assert "MAIN" in second.out


def test_recreate_commands_multiple(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """
[env_run_base]
package = "skip"
recreate_commands = [
    ["python", "-c", "print('CLEANUP_1')"],
    ["python", "-c", "print('CLEANUP_2')"],
]
commands = [["python", "-c", "print('MAIN')"]]
""",
    })
    first = proj.run("r")
    first.assert_success()

    second = proj.run("r", "-r")
    second.assert_success()
    assert "CLEANUP_1" in second.out
    assert "CLEANUP_2" in second.out
    assert "MAIN" in second.out


def test_recreate_commands_run_before_env_removed(tox_project: ToxProjectCreator) -> None:
    proj = tox_project({
        "tox.toml": """
[env_run_base]
package = "skip"
recreate_commands = [["python", "-c", "print('RECREATE_CLEANUP')"]]
commands = [["python", "-c", "print('MAIN')"]]
""",
    })
    first = proj.run("r")
    first.assert_success()

    second = proj.run("r", "-r")
    second.assert_success()
    lines = second.out.split("\n")
    cleanup_idx = next(i for i, line in enumerate(lines) if "RECREATE_CLEANUP" in line)
    remove_idx = next(i for i, line in enumerate(lines) if "remove tox env folder" in line)
    assert cleanup_idx < remove_idx


def test_package_only(
    tox_project: ToxProjectCreator,
    demo_pkg_inline: Path,
) -> None:
    ini = "[testenv]\ncommands = python -c 'print('foo')'"
    proj = tox_project(
        {"tox.ini": ini, "pyproject.toml": (demo_pkg_inline / "pyproject.toml").read_text()},
        base=demo_pkg_inline,
    )
    execute_calls = proj.patch_execute(lambda r: 0 if "install" in r.run_id else None)
    result = proj.run("r", "--sdistonly")
    result.assert_success()

    expected_calls = [
        (".pkg", "_optional_hooks"),
        (".pkg", "get_requires_for_build_sdist"),
        (".pkg", "get_requires_for_build_wheel"),
        (".pkg", "build_wheel"),
        (".pkg", "build_sdist"),
        (".pkg", "_exit"),
    ]
    found_calls = [(i[0][0].conf.name, i[0][3].run_id) for i in execute_calls.call_args_list]
    assert found_calls == expected_calls



from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING, Any

from tox.config.loader.native import to_native

if TYPE_CHECKING:
    from collections.abc import Callable

    from tox.config.sets import ConfigSet
    from tox.session.state import State


def build_structured_result(state: State) -> tuple[dict[str, Any], bool]:
    keys: list[str] = state.conf.options.list_keys_only
    show_everything = state.conf.options.env.is_all
    has_exception = False
    result: dict[str, Any] = {}

    envs: dict[str, Any] = {}
    for name in state.envs.iter(package=True):
        tox_env = state.envs[name]
        env_data, exc = _collect_conf(tox_env.conf, keys)
        if not keys:
            env_data = {"type": type(tox_env).__name__, **env_data}
        if exc:
            has_exception = True
        envs[tox_env.conf.name] = env_data
    result["env"] = envs

    if show_everything or state.conf.options.show_core:
        tox_data, exc = _collect_conf(state.conf.core, keys)
        has_exception = has_exception or exc
        result["tox"] = tox_data

    return result, has_exception


def write_output(
    output: str,
    output_file: Path | None,
    is_colored: bool,  # noqa: FBT001
    colorize: Callable[[str], str],
) -> None:
    if output_file is not None:
        Path(output_file).write_text(output + "\n", encoding="utf-8")
    else:
        print(colorize(output) if is_colored else output)  # noqa: T201


def _collect_conf(conf: ConfigSet, keys: list[str]) -> tuple[dict[str, Any], bool]:
    data: dict[str, Any] = {}
    has_exception = False
    for key in keys or conf:
        if key not in conf:
            continue
        key = conf.primary_key(key)  # noqa: PLW2901
        try:
            data[key] = to_native(conf[key])
        except Exception as exception:
            if os.environ.get("_TOX_SHOW_CONFIG_RAISE"):  # pragma: no branch
                raise  # pragma: no cover
            data[key] = {"error": repr(exception)}
            has_exception = True
    if (unused := conf.unused()) and not keys:
        data["unused"] = sorted(unused)
    return data, has_exception
